import asyncio

import typer

from gwproactor.command_line_utils import (
    print_settings,
    run_async_main,
)
from gwproactor_test.dummies import DUMMY_ATN_NAME
from gwproactor_test.dummies.tree.atn import DummyAtn
from gwproactor_test.dummies.tree.atn_settings import DummyAtnSettings

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    help="GridWorks Dummy Scada1",
)


@app.command()
def run(
    env_file: str = ".env",
    dry_run: bool = False,
    verbose: bool = False,
    message_summary: bool = False,
) -> None:
    asyncio.run(
        run_async_main(
            name=DUMMY_ATN_NAME,
            proactor_type=DummyAtn,
            settings_type=DummyAtnSettings,
            env_file=env_file,
            dry_run=dry_run,
            verbose=verbose,
            message_summary=message_summary,
        )
    )


@app.command()
def config(
    env_file: str = ".env",
) -> None:
    print_settings(settings_type=DummyAtnSettings, env_file=env_file)


@app.callback()
def _main() -> None: ...


if __name__ == "__main__":
    app()



import typing
from collections import defaultdict
from typing import Optional

import rich
from gwproto import Message

from gwproactor import Proactor
from gwproactor.links.link_settings import LinkSettings
from gwproactor.message import MQTTReceiptPayload
from gwproactor.persister import TimedRollingFilePersister
from gwproactor_test.dummies.names import DUMMY_SCADA2_SHORT_NAME
from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandSetRelay,
    AdminSetRelayEvent,
)
from gwproactor_test.dummies.tree.codecs import AdminCodec, DummyCodec
from gwproactor_test.dummies.tree.messages import (
    RelayReportEvent,
    SetRelay,
)
from gwproactor_test.dummies.tree.scada2_settings import DummyScada2Settings


class DummyScada2(Proactor):
    relays: dict[str, bool]

    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyScada2Settings] = None,
    ) -> None:
        if settings is None:
            settings = DummyScada2Settings()
        self.relays = defaultdict(bool)
        super().__init__(name=name, settings=settings)
        self._links.add_mqtt_link(
            LinkSettings(
                client_name=self.settings.scada1_link.client_name,
                gnode_name=self.settings.scada1_link.long_name,
                spaceheat_name=self.settings.scada1_link.short_name,
                mqtt=self.settings.scada1_link,
                codec=DummyCodec(
                    src_name=self.settings.scada1_link.long_name,
                    dst_name=DUMMY_SCADA2_SHORT_NAME,
                    model_name="Scada1ToScada2Message",
                ),
                upstream=True,
            ),
        )
        if self.settings.admin_link.enabled:
            self._links.add_mqtt_link(
                LinkSettings(
                    client_name=self.settings.admin_link.client_name,
                    gnode_name=self.settings.admin_link.long_name,
                    spaceheat_name=self.settings.admin_link.short_name,
                    mqtt=self.settings.admin_link,
                    codec=AdminCodec(),
                ),
            )
        self.links.log_subscriptions("construction")

    @property
    def settings(self) -> DummyScada2Settings:
        return typing.cast(DummyScada2Settings, self._settings)

    @property
    def subscription_name(self) -> str:
        return DUMMY_SCADA2_SHORT_NAME

    @property
    def admin_client(self) -> str:
        return self.settings.admin_link.client_name

    @classmethod
    def make_event_persister(
        cls, settings: DummyScada2Settings
    ) -> TimedRollingFilePersister:
        return TimedRollingFilePersister(settings.paths.event_dir)

    def _process_set_relay(self, payload: SetRelay) -> None:
        self._logger.path(
            f"++{self.name}._process_set_relay "
            f"{payload.RelayName}  "
            f"closed:{payload.Closed}"
        )
        path_dbg = 0
        last_val = self.relays[payload.RelayName]
        event = RelayReportEvent(
            relay_name=payload.RelayName,
            closed=payload.Closed,
            changed=last_val != payload.Closed,
        )
        if event.changed:
            path_dbg |= 0x00000001
            self.relays[payload.RelayName] = event.closed
        self.generate_event(event)
        self._logger.path(
            f"--{self.name}._process_set_relay  "
            f"path:0x{path_dbg:08X}  "
            f"{int(last_val)} -> "
            f"{int(event.closed)}"
        )

    def _process_upstream_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._process_downstream_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case SetRelay():
                path_dbg |= 0x00000001
                self._process_set_relay(decoded.Payload)
            case _:
                path_dbg |= 0x00000002
                rich.print(decoded.Header)
                raise ValueError(
                    f"There is no handler for mqtt message payload type [{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )
        self._logger.path(
            f"--{self.name}._process_downstream_mqtt_message  path:0x{path_dbg:08X}",
        )

    def _process_admin_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._process_admin_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case AdminCommandSetRelay() as command:
                path_dbg |= 0x00000001
                self.generate_event(AdminSetRelayEvent(command=command))
                self._process_set_relay(command.RelayInfo)
            case _:
                raise ValueError(
                    "In this test, since the environment is controlled, "
                    "there is no handler for mqtt message payload type "
                    f"[{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )

        self._logger.path(
            f"--{self.name}._process_admin_mqtt_message  path:0x{path_dbg:08X}",
        )

    def _derived_process_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._derived_process_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        if message.Payload.client_name == self.upstream_client:
            path_dbg |= 0x00000001
            self._process_upstream_mqtt_message(message, decoded)
        elif message.Payload.client_name == self.admin_client:
            path_dbg |= 0x00000002
            self._process_admin_mqtt_message(message, decoded)
        else:
            rich.print(decoded.Header)
            raise ValueError(
                "In this test, since the environment is controlled, "
                "there is no mqtt handler for message from client "
                f"[{message.Payload.client_name}]\n"
                f"Received\n\t topic: [{message.Payload.message.topic}]"
            )
        self._logger.path(
            f"--{self.name}._derived_process_mqtt_message  path:0x{path_dbg:08X}",
        )



import subprocess
import sys

import dotenv
import rich
import typer
from trogon import Trogon
from typer.main import get_group

from gwproactor_test import test_ca_certificate_path, test_ca_private_key_path
from gwproactor_test.certs import mqtt_client_fields
from gwproactor_test.dummies.tree import admin_cli, atn1_cli, scada1_cli, scada2_cli
from gwproactor_test.dummies.tree.admin_settings import DummyAdminSettings
from gwproactor_test.dummies.tree.atn_settings import DummyAtnSettings
from gwproactor_test.dummies.tree.scada1_settings import DummyScada1Settings
from gwproactor_test.dummies.tree.scada2_settings import DummyScada2Settings

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    help="GridWorks Proactor Test CLI",
)

app.add_typer(scada1_cli.app, name="scada1", help="Use dummy scada1")
app.add_typer(scada2_cli.app, name="scada2", help="Use dummy scada1")
app.add_typer(atn1_cli.app, name="atn", help="Use dummy scada1")
app.add_typer(admin_cli.app, name="admin", help="Use dummy admin")


@app.command()
def gen_dummy_certs(
    dry_run: bool = False,
    env_file: str = ".env",
) -> None:
    """Generate certs for dummy proactors."""
    console = rich.console.Console()
    env_file = dotenv.find_dotenv(env_file)
    ca_cert_path = test_ca_certificate_path()
    ca_private_key_path = test_ca_private_key_path()
    for settings in [
        DummyAtnSettings(_env_file=env_file),
        DummyScada1Settings(_env_file=env_file),
        DummyScada2Settings(_env_file=env_file),
        DummyAdminSettings(_env_file=env_file),
    ]:
        commands = []
        for _, client_config in mqtt_client_fields(settings):
            commands.append(
                [
                    "gwcert",
                    "key",
                    "add",
                    "--ca-certificate-path",
                    str(ca_cert_path),
                    "--ca-private-key-path",
                    str(ca_private_key_path),
                    "--force",
                    str(client_config.tls.paths.cert_path),
                ]
            )
        if dry_run:
            console.print(f"Showing key generation commands for {type(settings)}")
            for command in commands:
                console.print(f"  {' '.join(command)}", soft_wrap=True)
            console.print("\n\n\n")
        else:
            console.print(f"Generating keys with for {type(settings)}")
            for command in commands:
                console.print("Generating keys with command:")
                console.print(f"  {' '.join(command)}", soft_wrap=True)
                try:
                    result = subprocess.run(command, capture_output=True, check=True)  # noqa: S603
                    print(result.stdout.decode("utf-8"))  # noqa
                except subprocess.CalledProcessError as e:
                    print(e.stderr.decode("utf-8"))  # noqa
                    sys.exit(e.returncode)
            console.print("\n")


@app.command()
def commands(ctx: typer.Context) -> None:
    """CLI command builder."""
    Trogon(get_group(app), click_context=ctx).run()


@app.callback()
def main_app_callback() -> None: ...


# For sphinx:
typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()



from gwproto import MQTTCodec, create_message_model

from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandReadRelays,
    AdminCommandSetRelay,
)


class DummyCodec(MQTTCodec):
    src_name: str
    dst_name: str

    def __init__(self, src_name: str, dst_name: str, model_name: str) -> None:
        self.src_name = src_name
        self.dst_name = dst_name
        super().__init__(
            create_message_model(
                model_name=model_name,
                module_names=[
                    "gwproto.messages",
                    "gwproactor.message",
                    "gwproactor_test.dummies.tree.messages",
                ],
            )
        )

    def validate_source_and_destination(self, src: str, dst: str) -> None:
        if src != self.src_name or dst != self.dst_name:
            raise ValueError(
                "ERROR validating src and/or dst\n"
                f"  exp: {self.src_name} -> {self.dst_name}\n"
                f"  got: {src} -> {dst}"
            )


class AdminCodec(MQTTCodec):
    def __init__(self) -> None:
        super().__init__(
            create_message_model(
                model_name="AdminMessageDecoder",
                explicit_types=[AdminCommandSetRelay, AdminCommandReadRelays],
            )
        )

    def validate_source_and_destination(self, src: str, dst: str) -> None: ...



"""Scada implementation"""

from typing import Optional, cast

from gwproto import MQTTCodec, create_message_model

from gwproactor.links.link_settings import LinkSettings
from gwproactor.persister import SimpleDirectoryWriter
from gwproactor.proactor_implementation import Proactor
from gwproactor_test.dummies.names import (
    CHILD_SHORT_NAME,
    DUMMY_CHILD_NAME,
    DUMMY_PARENT_NAME,
    PARENT_SHORT_NAME,
)
from gwproactor_test.dummies.pair.parent_config import DummyParentSettings


class ParentMQTTCodec(MQTTCodec):
    def __init__(self) -> None:
        super().__init__(
            create_message_model(
                model_name="ParentMessageDecoder",
                module_names=["gwproto.messages", "gwproactor.message"],
            )
        )

    def validate_source_and_destination(self, src: str, dst: str) -> None:
        if src != DUMMY_CHILD_NAME or dst != PARENT_SHORT_NAME:
            raise ValueError(
                "ERROR validating src and/or dst\n"
                f"  exp: {DUMMY_CHILD_NAME} -> {PARENT_SHORT_NAME}\n"
                f"  got: {src} -> {dst}"
            )


class DummyParent(Proactor):
    CHILD_MQTT = "child"

    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyParentSettings] = None,
    ) -> None:
        super().__init__(
            name=name or DUMMY_PARENT_NAME,
            settings=DummyParentSettings() if settings is None else settings,
        )
        self._links.add_mqtt_link(
            LinkSettings(
                client_name=self.CHILD_MQTT,
                gnode_name=DUMMY_CHILD_NAME,
                spaceheat_name=CHILD_SHORT_NAME,
                mqtt=settings.child_mqtt,
                codec=ParentMQTTCodec(),
                downstream=True,
            )
        )

    @classmethod
    def make_event_persister(
        cls, settings: DummyParentSettings
    ) -> SimpleDirectoryWriter:
        return SimpleDirectoryWriter(settings.paths.event_dir)

    @property
    def publication_name(self) -> str:
        return self.name

    @property
    def settings(self) -> DummyParentSettings:
        return cast(DummyParentSettings, self._settings)

    @property
    def subscription_name(self) -> str:
        return PARENT_SHORT_NAME



from typing import Any, Self

from pydantic import model_validator
from pydantic_settings import SettingsConfigDict

from gwproactor import ProactorSettings
from gwproactor_test.dummies import (
    DUMMY_ATN_NAME,
    DUMMY_SCADA1_ENV_PREFIX,
    DUMMY_SCADA1_NAME,
    DUMMY_SCADA2_NAME,
)
from gwproactor_test.dummies.names import DUMMY_ATN_SHORT_NAME, DUMMY_SCADA2_SHORT_NAME
from gwproactor_test.dummies.tree.admin_settings import AdminLinkSettings
from gwproactor_test.dummies.tree.link_settings import TreeLinkSettings


class AtnLinkSettings(TreeLinkSettings):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            client_name=DUMMY_ATN_NAME,
            long_name=DUMMY_ATN_NAME,
            short_name=DUMMY_ATN_SHORT_NAME,
            **kwargs,
        )


class Scada2LinkSettings(TreeLinkSettings):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            client_name=DUMMY_SCADA2_NAME,
            long_name=DUMMY_SCADA2_NAME,
            short_name=DUMMY_SCADA2_SHORT_NAME,
            **kwargs,
        )


class DummyScada1Settings(ProactorSettings):
    atn_link: AtnLinkSettings = AtnLinkSettings()
    scada2_link: Scada2LinkSettings = Scada2LinkSettings()
    admin_link: AdminLinkSettings = AdminLinkSettings()

    model_config = SettingsConfigDict(env_prefix=DUMMY_SCADA1_ENV_PREFIX)

    @model_validator(mode="before")
    @classmethod
    def pre_root_validator(cls, values: dict) -> dict:
        return ProactorSettings.update_paths_name(values, DUMMY_SCADA1_NAME)

    @model_validator(mode="after")
    def validate(self) -> Self:
        self.atn_link.update_tls_paths(self.paths.certs_dir, self.atn_link.client_name)
        self.scada2_link.update_tls_paths(
            self.paths.certs_dir, self.scada2_link.client_name
        )
        return self



from enum import StrEnum

import rich
import typer

from gwproactor.command_line_utils import get_settings, print_settings
from gwproactor_test.dummies.tree.admin import MQTTAdmin
from gwproactor_test.dummies.tree.admin_settings import (
    DummyAdminSettings,
)

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    help="GridWorks Dummy Admin Client",
)


class RelayState(StrEnum):
    open = "0"
    closed = "1"


def _set_relay(
    *,
    target: str,
    relay_name: str,
    closed: RelayState,
    user: str = "HeatpumpWizard",
    json: bool = False,
) -> None:
    settings = DummyAdminSettings(target_gnode=target)
    if not json:
        rich.print(settings)
    admin = MQTTAdmin(
        settings=settings,
        relay_name=relay_name,
        closed=closed == RelayState.closed,
        user=user,
        json=json,
    )
    admin.run()


@app.command()
def set_relay(
    target: str,
    relay_name: str,
    closed: RelayState,
    user: str = "HeatpumpWizard",
    json: bool = False,
) -> None:
    _set_relay(
        target=target,
        relay_name=relay_name,
        closed=closed,
        user=user,
        json=json,
    )


@app.command()
def run(
    target: str = "dummy_scada1",
    relay_name: str = "relay0",
    closed: RelayState = RelayState.closed,
    user: str = "HeatpumpWizard",
    json: bool = False,
) -> None:
    _set_relay(
        target=target,
        relay_name=relay_name,
        closed=closed,
        user=user,
        json=json,
    )


@app.command()
def config(
    target: str = "",
    env_file: str = ".env",
) -> None:
    settings = get_settings(settings_type=DummyAdminSettings, env_file=env_file)
    settings.target_gnode = target
    print_settings(settings=settings, env_file=env_file)


@app.callback()
def _main() -> None: ...


if __name__ == "__main__":
    app()



from gwproactor_test.dummies.names import (
    DUMMY_ATN_ENV_PREFIX,
    DUMMY_ATN_NAME,
    DUMMY_CHILD_ENV_PREFIX,
    DUMMY_CHILD_NAME,
    DUMMY_ENV_PREFIX,
    DUMMY_PARENT_ENV_PREFIX,
    DUMMY_PARENT_NAME,
    DUMMY_SCADA1_ENV_PREFIX,
    DUMMY_SCADA1_NAME,
    DUMMY_SCADA2_ENV_PREFIX,
    DUMMY_SCADA2_NAME,
)
from gwproactor_test.dummies.pair.child import DummyChild
from gwproactor_test.dummies.pair.child_config import DummyChildSettings
from gwproactor_test.dummies.pair.parent import DummyParent
from gwproactor_test.dummies.pair.parent_config import DummyParentSettings

__all__ = [
    "DUMMY_CHILD_ENV_PREFIX",
    "DUMMY_CHILD_NAME",
    "DUMMY_ENV_PREFIX",
    "DUMMY_PARENT_ENV_PREFIX",
    "DUMMY_PARENT_NAME",
    "DummyChild",
    "DummyChildSettings",
    "DummyParent",
    "DummyParentSettings",
    "DUMMY_SCADA1_ENV_PREFIX",
    "DUMMY_SCADA2_ENV_PREFIX",
    "DUMMY_SCADA1_NAME",
    "DUMMY_SCADA2_NAME",
    "DUMMY_ATN_NAME",
    "DUMMY_ATN_ENV_PREFIX",
]



import contextlib
import datetime
import logging
from typing import Any, Optional, Sequence, TypeAlias


class MessageSummary:
    """Helper class for formating message summaries message receipt/publication single line summaries."""

    DEFAULT_FORMAT = (
        "  {direction:15s}  {src_dst:50s}  {broker_flag}  {arrow:2s}  {topic:80s}"
        "  {payload_type:25s}{message_id}"
    )

    @classmethod
    def format(  # noqa: PLR0913
        cls,
        direction: str,
        src: str,
        dst: str,
        topic: str,
        *,
        payload_object: Any = None,
        broker_flag: str = " ",
        timestamp: Optional[datetime.datetime] = None,
        include_timestamp: bool = False,
        message_id: str = "",
    ) -> str:
        """
        Formats a single line summary of message receipt/publication.

        Args:
            direction: "IN" or "OUT"
            src: The node name of the sending or receiving actor.
            dst: The node name of the inteded recipient
            topic: The destination or source topic.
            payload_object: The payload of the message.
            broker_flag: "*" for the "gw" broker.
            timestamp: datetime.dateime.now(tz=datetime.timezone.utc) by default.
            include_timestamp: whether timestamp is prepended to output.
            message_id: The message id. Ignored if empty.

        Returns:
            Formatted string.
        """
        with contextlib.suppress(Exception):
            if timestamp is None:
                timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
            if include_timestamp:
                format_ = "{timestamp}  " + cls.DEFAULT_FORMAT
            else:
                format_ = cls.DEFAULT_FORMAT
            direction = direction.strip()
            if direction.startswith(("OUT", "SND")):
                arrow = "->"
            elif direction.startswith(("IN", "RCV")):  # noqa: PIE810
                arrow = "<-"
            else:
                arrow = "? "
            if hasattr(payload_object, "payload"):
                payload_object = payload_object.payload
            if hasattr(payload_object, "__class__"):
                payload_str = payload_object.__class__.__name__
            else:
                payload_str = type(payload_object)
            max_msg_id_len = 11
            if message_id and len(message_id) > max_msg_id_len:
                message_id = f" {message_id[:max_msg_id_len-3]}..."
            src_dst = f"{src} to {dst}"
            return format_.format(
                timestamp=timestamp.isoformat(),
                direction=direction,
                src_dst=src_dst,
                broker_flag=broker_flag,
                arrow=arrow,
                topic=f"[{topic}]",
                payload_type=payload_str,
                message_id=message_id,
            )
        return ""


LoggerOrAdapter: TypeAlias = logging.Logger | logging.LoggerAdapter


class CategoryLoggerInfo:
    logger: LoggerOrAdapter
    default_level: int = logging.INFO
    default_disabled = False

    def __init__(
        self, logger: LoggerOrAdapter, default_level: int = logging.INFO
    ) -> None:
        self.logger = logger
        self.default_level = default_level
        self.default_disabled = logger.disabled


class ProactorLogger(logging.LoggerAdapter):
    MESSAGE_DELIMITER_WIDTH = 88
    MESSAGE_ENTRY_DELIMITER = "+" * MESSAGE_DELIMITER_WIDTH
    MESSAGE_EXIT_DELIMITER = "-" * MESSAGE_DELIMITER_WIDTH

    message_summary_logger: logging.Logger
    lifecycle_logger: logging.Logger
    comm_event_logger: logging.Logger
    category_loggers: dict[str, CategoryLoggerInfo]

    def __init__(  # noqa: PLR0913
        self,
        base: str,
        message_summary: str,
        lifecycle: str,
        comm_event: str,
        extra: Optional[dict] = None,
        category_logger_names: Optional[Sequence[str]] = None,
    ) -> None:
        super().__init__(logging.getLogger(base), extra=extra)
        self.message_summary_logger = logging.getLogger(message_summary)
        self.lifecycle_logger = logging.getLogger(lifecycle)
        self.comm_event_logger = logging.getLogger(comm_event)
        self.category_loggers = {}
        if category_logger_names is not None:
            for category_logger_name in category_logger_names:
                self.add_category_logger(category_logger_name)

    @property
    def general_enabled(self) -> bool:
        return self.logger.isEnabledFor(logging.INFO)

    @property
    def message_summary_enabled(self) -> bool:
        return self.message_summary_logger.isEnabledFor(logging.INFO)

    @property
    def path_enabled(self) -> bool:
        return self.message_summary_logger.isEnabledFor(logging.DEBUG)

    @property
    def lifecycle_enabled(self) -> bool:
        return self.lifecycle_logger.isEnabledFor(logging.INFO)

    @property
    def comm_event_enabled(self) -> bool:
        return self.comm_event_logger.isEnabledFor(logging.INFO)

    def message_summary(  # noqa: PLR0913
        self,
        *,
        direction: str,
        src: str,
        dst: str,
        topic: str,
        payload_object: Any = None,
        broker_flag: str = " ",
        timestamp: Optional[datetime.datetime] = None,
        message_id: str = "",
    ) -> None:
        if self.message_summary_logger.isEnabledFor(logging.INFO):
            self.message_summary_logger.info(
                MessageSummary.format(
                    direction=direction,
                    src=src,
                    dst=dst,
                    topic=topic,
                    payload_object=payload_object,
                    broker_flag=broker_flag,
                    timestamp=timestamp,
                    message_id=message_id,
                )
            )

    def path(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.message_summary_logger.debug(msg, *args, **kwargs)

    def lifecycle(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.lifecycle_logger.info(msg, *args, **kwargs)

    def comm_event(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.comm_event_logger.info(msg, *args, **kwargs)

    def message_enter(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.path_enabled:
            self.path("")
            self.path(self.MESSAGE_ENTRY_DELIMITER)
            self.path(msg, *args, **kwargs)

    def message_exit(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.path_enabled:
            self.path(msg, *args, **kwargs)
            self.path(self.MESSAGE_EXIT_DELIMITER)

    def category_logger_name(self, category: str) -> str:
        return f"{self.name}.{category}"

    def category_logger(
        self, category: str
    ) -> Optional[logging.Logger | logging.LoggerAdapter]:
        """Get existing category logger"""
        logger_info = self.category_loggers.get(category)
        if logger_info is not None:
            return logger_info.logger
        return None

    def add_category_logger(
        self,
        category: str = "",
        level: int = logging.INFO,
        logger: Optional[LoggerOrAdapter] = None,
    ) -> LoggerOrAdapter:
        if logger is None:
            if not category:
                raise ValueError(
                    "ERROR. add_category_logger() requires category value "
                    "unless logger is provided."
                )
            logger = self.category_logger(category)
            if logger is None:
                logger = logging.getLogger(self.category_logger_name(category))
                logger.setLevel(level)
                self.category_loggers[category] = CategoryLoggerInfo(
                    logger=logger,
                    default_level=level,
                )
        else:
            if not category:
                category = logger.name
                self_prefix = f"{self.name}."
                if category.startswith(self_prefix):
                    category = category[len(self_prefix) :]
            if category in self.category_loggers:
                raise ValueError(
                    "ERROR. add_category_logger() got explicit logger "
                    f"named {logger.name}, categorized as {category}, but "
                    f"logger for that category is already present."
                )
            level = logger.getEffectiveLevel()
            self.category_loggers[category] = CategoryLoggerInfo(
                logger=logger,
                default_level=level,
            )
        return logger

    def reset_default_category_levels(self) -> None:
        for logger_info in self.category_loggers.values():
            logger_info.logger.setLevel(logger_info.default_level)
            logger_info.logger.disabled = logger_info.default_disabled

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"{self.logger.name}, "
            f"{self.message_summary_logger.name}, "
            f"{self.lifecycle_logger.name}, "
            f"{self.comm_event_logger.name}>"
        )



from gwproto import Message, MQTTCodec, MQTTTopic

from gwproactor.config import MQTTClient


class LinkSettings:
    client_name: str
    gnode_name: str
    spaceheat_name: str
    subscription_name: str = ""
    mqtt: MQTTClient
    codec: MQTTCodec
    upstream: bool = False
    downstream: bool = False

    def __init__(  # noqa: PLR0913
        self,
        *,
        client_name: str,
        gnode_name: str,
        spaceheat_name: str,
        mqtt: MQTTClient,
        codec: MQTTCodec,
        subscription_name: str = "",
        upstream: bool = False,
        downstream: bool = False,
    ) -> None:
        if upstream and downstream:
            raise ValueError(
                f"ERROR. Link {client_name} can be 0 or 1 of upstream and "
                "downstream, but not both."
            )
        self.client_name = client_name
        self.gnode_name = gnode_name
        self.spaceheat_name = spaceheat_name
        self.subscription_name = subscription_name
        self.upstream = upstream
        self.mqtt = mqtt
        self.codec = codec
        self.downstream = downstream

    def subscription_topic(self, receiver_spaceheat_name: str) -> str:
        return MQTTTopic.encode(
            envelope_type=Message.type_name(),
            src=self.gnode_name,
            dst=receiver_spaceheat_name,
            message_type="#",
        )



from typing import Self

from pydantic import BaseModel, Field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from gwproactor.config.logging import LoggingSettings
from gwproactor.config.mqtt import MQTTClient
from gwproactor.config.paths import Paths

MQTT_LINK_POLL_SECONDS = 60.0
ACK_TIMEOUT_SECONDS = 5.0
NUM_INITIAL_EVENT_REUPLOADS: int = 5


class ProactorSettings(BaseSettings):
    paths: Paths = Field({}, validate_default=True)
    logging: LoggingSettings = LoggingSettings()
    mqtt_link_poll_seconds: float = MQTT_LINK_POLL_SECONDS
    ack_timeout_seconds: float = ACK_TIMEOUT_SECONDS
    num_initial_event_reuploads: int = NUM_INITIAL_EVENT_REUPLOADS
    model_config = SettingsConfigDict(env_prefix="PROACTOR_", env_nested_delimiter="__")

    @field_validator("paths")
    @classmethod
    def get_paths(cls, v: Paths) -> Paths:
        if not v:
            v = Paths()
        return v

    @classmethod
    def update_paths_name(cls, values: dict, name: str) -> dict:
        """Update paths member with a new 'name' attribute, e.g., a name known by a derived class.

        This is meant to be called in a 'pre=True' root validator of a derived class.
        """
        if "paths" not in values:
            values["paths"] = Paths(name=name)
        elif isinstance(values["paths"], BaseModel):
            if "name" not in values["paths"].model_fields_set:
                values["paths"] = values["paths"].copy(name=name, deep=True)
        elif "name" not in values["paths"]:
            values["paths"]["name"] = name
        return values

    @model_validator(mode="after")
    def post_root_validator(self) -> Self:
        """Update unset paths of any member MQTTClient's TLS paths based on ProactorSettings 'paths' member."""
        if not isinstance(self.paths, Paths):
            raise ValueError(  # noqa: TRY004
                f"ERROR. 'paths' member must be instance of Paths. Got: {type(self.paths)}"
            )
        for field_name in self.model_fields:
            v = getattr(self, field_name)
            if isinstance(v, MQTTClient):
                v.update_tls_paths(self.paths.certs_dir, field_name)
        return self



import argparse
import asyncio
import contextlib
import logging
from dataclasses import dataclass, field
from pathlib import Path
from types import TracebackType
from typing import Callable, Optional, Type, TypeVar

from gwproactor import Proactor, ProactorSettings, setup_logging
from gwproactor.config import DEFAULT_BASE_NAME, LoggingSettings, MQTTClient, Paths
from gwproactor_test import copy_keys
from gwproactor_test.certs import uses_tls
from gwproactor_test.logger_guard import LoggerGuards
from gwproactor_test.recorder import (
    ProactorT,
    RecorderInterface,
    make_recorder_class,
)


@dataclass
class ProactorTestHelper:
    name: str
    path_name: str
    settings: ProactorSettings
    kwargs: dict = field(default_factory=dict)
    proactor: Optional[RecorderInterface] = None


ChildT = TypeVar("ChildT", bound=Proactor)
ParentT = TypeVar("ParentT", bound=Proactor)
ChildSettingsT = TypeVar("ChildSettingsT", bound=ProactorSettings)
ParentSettingsT = TypeVar("ParentSettingsT", bound=ProactorSettings)


class CommTestHelper:
    parent_t: Type[ProactorT]
    child_t: Type[Proactor]
    parent_settings_t: Type[ParentSettingsT]
    child_settings_t: Type[ProactorSettings]

    parent_recorder_t: Callable[..., RecorderInterface] = None
    child_recorder_t: Callable[..., RecorderInterface] = None

    parent_helper: ProactorTestHelper
    child_helper: ProactorTestHelper
    verbose: bool
    child_verbose: bool
    parent_verbose: bool
    parent_on_screen: bool
    lifecycle_logging: bool
    logger_guards: LoggerGuards

    warn_if_multi_subscription_tests_skipped: bool = True

    @classmethod
    def setup_class(cls) -> None:
        if cls.parent_recorder_t is None:
            cls.parent_recorder_t = make_recorder_class(cls.parent_t)
        if cls.child_recorder_t is None:
            cls.child_recorder_t = make_recorder_class(cls.child_t)

    def __init__(
        self,
        *,
        child_settings: Optional[ChildSettingsT] = None,
        parent_settings: Optional[ParentSettingsT] = None,
        verbose: bool = False,
        child_verbose: bool = False,
        parent_verbose: bool = False,
        lifecycle_logging: bool = False,
        add_child: bool = False,
        add_parent: bool = False,
        start_child: bool = False,
        start_parent: bool = False,
        parent_on_screen: bool = False,
        child_name: str = "",
        parent_name: str = "",
        child_path_name: str = "child",
        parent_path_name: str = "parent",
        child_kwargs: Optional[dict] = None,
        parent_kwargs: Optional[dict] = None,
    ) -> None:
        self.setup_class()
        self.child_helper = ProactorTestHelper(
            child_name,
            child_path_name,
            (
                self.child_settings_t(
                    logging=LoggingSettings(
                        base_log_name=f"{child_path_name}_{DEFAULT_BASE_NAME}"
                    ),
                    paths=Paths(name=Path(child_path_name)),
                )
                if child_settings is None
                else child_settings
            ),
            {} if child_kwargs is None else child_kwargs,
        )
        self.parent_helper = ProactorTestHelper(
            parent_name,
            parent_path_name,
            (
                self.parent_settings_t(
                    logging=LoggingSettings(
                        base_log_name=f"{parent_path_name}_{DEFAULT_BASE_NAME}"
                    ),
                    paths=Paths(name=Path(parent_path_name)),
                )
                if parent_settings is None
                else parent_settings
            ),
            {} if parent_kwargs is None else parent_kwargs,
        )
        self.verbose = verbose
        self.child_verbose = child_verbose
        self.parent_verbose = parent_verbose
        self.parent_on_screen = parent_on_screen
        self.lifecycle_logging = lifecycle_logging
        self.setup_logging()
        if add_child or start_child:
            self.add_child()
            if start_child:
                self.start_child()
        if add_parent or start_parent:
            self.add_parent()
            if start_parent:
                self.start_parent()

    @classmethod
    def _make(
        cls, recorder_t: Callable[..., RecorderInterface], helper: ProactorTestHelper
    ) -> RecorderInterface:
        if uses_tls(helper.settings):
            copy_keys(helper.path_name, helper.settings)
        return recorder_t(helper.name, helper.settings, **helper.kwargs)

    def make_parent(self) -> RecorderInterface:
        return self._make(self.parent_recorder_t, self.parent_helper)

    def make_child(self) -> RecorderInterface:
        return self._make(self.child_recorder_t, self.child_helper)

    @property
    def parent(self) -> Optional[ProactorT]:
        return self.parent_helper.proactor

    @property
    def child(self) -> Optional[ProactorT]:
        return self.child_helper.proactor

    def start_child(self) -> "CommTestHelper":
        if self.child is not None:
            self.start_proactor(self.child)
        return self

    def start_parent(self) -> "CommTestHelper":
        if self.parent is not None:
            self.start_proactor(self.parent)
        return self

    def start_proactor(self, proactor: Proactor) -> "CommTestHelper":
        asyncio.create_task(proactor.run_forever(), name=f"{proactor.name}_run_forever")  # noqa: RUF006
        return self

    def start(self) -> "CommTestHelper":
        return self

    def add_child(self) -> "CommTestHelper":
        self.child_helper.proactor = self.make_child()
        return self

    def add_parent(self) -> "CommTestHelper":
        self.parent_helper.proactor = self.make_parent()
        return self

    def remove_child(self) -> "CommTestHelper":
        self.child_helper.proactor = None
        return self

    def remove_parent(self) -> "CommTestHelper":
        self.parent_helper.proactor = None
        return self

    @classmethod
    def _get_clients_supporting_tls(
        cls, settings: ProactorSettings
    ) -> list[MQTTClient]:
        clients = []
        for field_name in settings.model_fields:
            v = getattr(settings, field_name)
            if isinstance(v, MQTTClient):
                clients.append(v)
        return clients

    def _get_child_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ChildSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.child_helper.settings)

    def _get_parent_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ParentSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.parent_helper.settings)

    @classmethod
    def _set_settings_use_tls(cls, use_tls: bool, clients: list[MQTTClient]) -> None:
        for client in clients:
            client.tls.use_tls = use_tls

    def set_use_tls(self, use_tls: bool) -> None:
        """Set MQTTClients which support TLS in parent and child settings to use TLS per use_tls. Clients supporting TLS
        is determined by _get_child_clients_supporting_tls() and _get_parent_clients_supporting_tls() which may be
        overriden in derived class.
        """
        self._set_settings_use_tls(use_tls, self._get_child_clients_supporting_tls())
        self._set_settings_use_tls(use_tls, self._get_parent_clients_supporting_tls())

    def setup_logging(self) -> None:
        self.child_helper.settings.paths.mkdirs(parents=True)
        self.parent_helper.settings.paths.mkdirs(parents=True)
        errors = []
        if not self.lifecycle_logging and not self.verbose:
            if not self.child_verbose:
                self.child_helper.settings.logging.levels.lifecycle = logging.WARNING
            if not self.parent_verbose:
                self.parent_helper.settings.logging.levels.lifecycle = logging.WARNING
        self.logger_guards = LoggerGuards(
            list(self.child_helper.settings.logging.qualified_logger_names().values())
            + list(
                self.parent_helper.settings.logging.qualified_logger_names().values()
            )
        )
        setup_logging(
            argparse.Namespace(verbose=self.verbose or self.child_verbose),
            self.child_helper.settings,
            errors=errors,
            add_screen_handler=True,
            root_gets_handlers=False,
        )
        assert not errors
        setup_logging(
            argparse.Namespace(verbose=self.verbose or self.parent_verbose),
            self.parent_helper.settings,
            errors=errors,
            add_screen_handler=self.parent_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    def get_proactors(self) -> list[RecorderInterface]:
        return [
            helper.proactor
            for helper in [self.child_helper, self.parent_helper]
            if helper.proactor is not None
        ]

    async def stop_and_join(self) -> None:
        proactors = self.get_proactors()
        for proactor in proactors:
            with contextlib.suppress(Exception):
                proactor.stop()
        for proactor in proactors:
            with contextlib.suppress(Exception):
                await proactor.join()

    async def __aenter__(self) -> "CommTestHelper":
        return self

    def get_log_path_str(self, exc: BaseException) -> str:
        return (
            f"CommTestHelper caught error {exc}.\n"
            "Working log dirs:"
            f"\n\t[{self.child_helper.settings.paths.log_dir}]"
            f"\n\t[{self.parent_helper.settings.paths.log_dir}]"
        )

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        try:
            await self.stop_and_join()
        finally:
            if exc is not None:
                try:
                    s = self.get_log_path_str(exc)
                except Exception as e:  # noqa: BLE001
                    try:
                        s = (
                            f"Caught {type(e)} / <{e}> while logging "
                            f"{type(exc)} / <{exc}>"
                        )
                    except:  # noqa: E722
                        s = "ERRORs upon errors in CommTestHelper cleanup"
                with contextlib.suppress(Exception):
                    logging.getLogger("gridworks").error(s)
            with contextlib.suppress(Exception):
                self.logger_guards.restore()
        return False

    def summary_str(self) -> str:
        s = ""
        if self.child:
            s += "CHILD:\n" f"{self.child.summary_str()}\n"
        else:
            s += "CHILD: None\n"
        if self.parent:
            s += "PARENT:\n" f"{self.parent.summary_str()}"
        else:
            s += "PARENT: None\n"
        return s



DUMMY_PARENT_NAME = "parent"
DUMMY_CHILD_NAME = "child"
PARENT_SHORT_NAME = "p"
CHILD_SHORT_NAME = "c"
DUMMY_ENV_PREFIX = "GW"
DUMMY_ATN_NAME = "dummy_atn1"
DUMMY_SCADA1_NAME = "dummy_scada1"
DUMMY_SCADA2_NAME = "dummy_scada2"
DUMMY_ADMIN_NAME = "dummy_admin"
DUMMY_ATN_SHORT_NAME = "a"
DUMMY_SCADA1_SHORT_NAME = "s"
DUMMY_SCADA2_SHORT_NAME = "s2"
DUMMY_ADMIN_SHORT_NAME = "dummy_admin"
DUMMY_CHILD_ENV_PREFIX = DUMMY_ENV_PREFIX + "CHILD_"
DUMMY_PARENT_ENV_PREFIX = DUMMY_ENV_PREFIX + "PARENT_"
DUMMY_ATN_ENV_PREFIX = DUMMY_ENV_PREFIX + "ATN1_"
DUMMY_SCADA1_ENV_PREFIX = DUMMY_ENV_PREFIX + "SCADA1_"
DUMMY_SCADA2_ENV_PREFIX = DUMMY_ENV_PREFIX + "SCADA1_"



from pydantic import model_validator
from pydantic_settings import SettingsConfigDict

from gwproactor import ProactorSettings
from gwproactor.config import MQTTClient
from gwproactor_test.dummies.names import DUMMY_CHILD_ENV_PREFIX, DUMMY_CHILD_NAME


class DummyChildSettings(ProactorSettings):
    parent_mqtt: MQTTClient = MQTTClient()
    seconds_per_report: int = 300
    async_power_reporting_threshold: float = 0.02

    model_config = SettingsConfigDict(env_prefix=DUMMY_CHILD_ENV_PREFIX)

    @model_validator(mode="before")
    @classmethod
    def pre_root_validator(cls, values: dict) -> dict:
        return ProactorSettings.update_paths_name(values, DUMMY_CHILD_NAME)



import typing
from typing import Optional

import rich
from gwproto import Message
from gwproto.messages import EventBase

from gwproactor import Proactor
from gwproactor.links.link_settings import LinkSettings
from gwproactor.message import MQTTReceiptPayload
from gwproactor.persister import TimedRollingFilePersister
from gwproactor_test.dummies.names import DUMMY_SCADA1_SHORT_NAME
from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandReadRelays,
    AdminCommandSetRelay,
    AdminSetRelayEvent,
)
from gwproactor_test.dummies.tree.codecs import AdminCodec, DummyCodec
from gwproactor_test.dummies.tree.messages import (
    RelayInfoReported,
    RelayReportEvent,
    RelayReportReceivedEvent,
    RelayStates,
    SetRelay,
    SetRelayMessage,
)
from gwproactor_test.dummies.tree.scada1_settings import DummyScada1Settings


class DummyScada1(Proactor):
    relays: RelayStates

    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyScada1Settings] = None,
    ) -> None:
        self.relays = RelayStates()
        if settings is None:
            settings = DummyScada1Settings()
        super().__init__(name=name, settings=settings)

        self._links.add_mqtt_link(
            LinkSettings(
                client_name=self.settings.atn_link.client_name,
                gnode_name=self.settings.atn_link.long_name,
                spaceheat_name=self.settings.atn_link.short_name,
                mqtt=self.settings.atn_link,
                codec=DummyCodec(
                    src_name=self.settings.atn_link.long_name,
                    dst_name=DUMMY_SCADA1_SHORT_NAME,
                    model_name="Atn1ToScada1Message",
                ),
                upstream=True,
            ),
        )
        self._links.add_mqtt_link(
            LinkSettings(
                client_name=self.settings.scada2_link.client_name,
                gnode_name=self.settings.scada2_link.long_name,
                spaceheat_name=self.settings.scada2_link.short_name,
                mqtt=self.settings.scada2_link,
                codec=DummyCodec(
                    src_name=self.settings.scada2_link.long_name,
                    dst_name=DUMMY_SCADA1_SHORT_NAME,
                    model_name="Scada2ToScada1Message",
                ),
                downstream=True,
            ),
        )
        if self.settings.admin_link.enabled:
            self._links.add_mqtt_link(
                LinkSettings(
                    client_name=self.settings.admin_link.client_name,
                    gnode_name=self.settings.admin_link.long_name,
                    spaceheat_name=self.settings.admin_link.long_name,
                    subscription_name=self.publication_name,
                    mqtt=self.settings.admin_link,
                    codec=AdminCodec(),
                ),
            )
        self.links.log_subscriptions("construction")
        self.links.enable_mqtt_loggers(self.logger.message_summary_logger)

    @property
    def settings(self) -> DummyScada1Settings:
        return typing.cast(DummyScada1Settings, self._settings)

    @property
    def admin_client(self) -> str:
        return self.settings.admin_link.client_name

    @property
    def subscription_name(self) -> str:
        return DUMMY_SCADA1_SHORT_NAME

    @classmethod
    def make_event_persister(
        cls, settings: DummyScada1Settings
    ) -> TimedRollingFilePersister:
        return TimedRollingFilePersister(settings.paths.event_dir)

    def set_relay(self, relay_name: str, closed: bool) -> None:
        self.send_threadsafe(
            SetRelayMessage(
                src=self.name,
                dst=self.name,
                relay_name=relay_name,
                closed=closed,
            )
        )

    def _derived_process_message(self, message: Message) -> None:
        self._logger.path(
            f"++{self.name}._derived_process_message "
            f"{message.Header.Src}/{message.Header.MessageType}"
        )
        path_dbg = 0
        match message.Payload:
            case SetRelay():
                path_dbg |= 0x00000001
                self._links.publish_message(
                    self.downstream_client,
                    SetRelayMessage(
                        src=self.publication_name,
                        relay_name=message.Payload.RelayName,
                        closed=message.Payload.Closed,
                        ack_required=True,
                    ),
                )
            case _:
                path_dbg |= 0x00000002
        self._logger.path(
            "--{self.name}._derived_process_message  path:0x{path_dbg:08X}"
        )

    def _process_report_relay_event(self, event: RelayReportEvent) -> None:
        self._logger.path(
            f"++{self.name}._process_set_relay_event "
            f"{event.relay_name}  "
            f"closed:{event.closed}  "
            f"changed: {event.changed}"
        )
        path_dbg = 0
        if event.relay_name not in self.relays.Relays:
            self.relays.Relays[event.relay_name] = RelayInfoReported()
        last_val = self.relays.Relays[event.relay_name].Closed
        self.relays.Relays[event.relay_name].Closed = event.closed
        changed = last_val != self.relays.Relays[event.relay_name].Closed
        self.logger.info(
            f"{event.relay_name}:  {int(last_val)} -> "
            f"{int(self.relays.Relays[event.relay_name].Closed)}  "
            f"changed: {int(changed)}/{int(event.changed)}"
        )
        report_received_event = RelayReportReceivedEvent(
            relay_name=event.relay_name,
            closed=event.closed,
            changed=event.changed,
        )

        if changed != event.changed:
            path_dbg |= 0x00000001
            report_received_event.mismatch = True
            self.relays.Relays[event.relay_name].CurrentChangeMismatch = True
            self.relays.Relays[event.relay_name].MismatchCount += 1
            self.relays.TotalChangeMismatches += 1
            report_received_event.mismatch_count = self.relays.TotalChangeMismatches
            self.logger.info(
                f"State change mismatch for {event.relay_name}  "
                f"found: {int(changed)}  reported: {event.changed}  "
                f"total mismatches: {self.relays.TotalChangeMismatches}"
            )
        self.generate_event(report_received_event)
        self._logger.path(
            f"--{self.name}._process_set_relay_event "
            f"{event.relay_name}  "
            f"closed:{event.closed}  "
            f"changed: {event.changed}  "
            f"path: 0x{path_dbg:08X}"
        )

    def _process_event(self, event: EventBase) -> None:
        self._logger.path(
            f"++_process_event  {event.TypeName}  from:{event.Src}",
        )
        self.generate_event(event)
        if isinstance(event, RelayReportEvent):
            self._process_report_relay_event(event)
        self._logger.path("--_process_event")

    def _process_downstream_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._process_downstream_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case EventBase():
                path_dbg |= 0x00000001
                self._process_event(decoded.Payload)
            case _:
                # For testing purposes, this should fail.
                path_dbg |= 0x00000002
                rich.print(decoded.Header)
                raise ValueError(
                    "In this test, since the environment is controlled, "
                    "there is no handler for mqtt message payload type "
                    f"[{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )
        self._logger.path(
            f"--{self.name}._process_downstream_mqtt_message  path:0x{path_dbg:08X}",
        )

    def _process_admin_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._process_admin_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case AdminCommandSetRelay() as command:
                path_dbg |= 0x00000001
                self.generate_event(AdminSetRelayEvent(command=command))
                self._links.publish_message(
                    self.downstream_client,
                    SetRelayMessage(
                        src=self.publication_name,
                        relay_name=command.RelayInfo.RelayName,
                        closed=command.RelayInfo.Closed,
                        ack_required=True,
                    ),
                )
            case AdminCommandReadRelays():
                path_dbg |= 0x00000002
                self._links.publish_message(
                    self.admin_client,
                    Message(
                        Src=self.publication_name, Payload=self.relays.model_copy()
                    ),
                )
            case _:
                raise ValueError(
                    "In this test, since the environment is controlled, "
                    "there is no handler for mqtt message payload type "
                    f"[{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )

        self._logger.path(
            f"--{self.name}._process_admin_mqtt_message  path:0x{path_dbg:08X}",
        )

    def _derived_process_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._derived_process_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        if message.Payload.client_name == self.downstream_client:
            path_dbg |= 0x00000001
            self._process_downstream_mqtt_message(message, decoded)
        elif message.Payload.client_name == self.admin_client:
            path_dbg |= 0x00000002
            self._process_admin_mqtt_message(message, decoded)
        else:
            rich.print(decoded.Header)
            raise ValueError(
                "In this test, since the environment is controlled, "
                "there is no mqtt handler for message from client "
                f"[{message.Payload.client_name}]\n"
                f"Received\n\t topic: [{message.Payload.message.topic}]"
            )
        self._logger.path(
            f"--{self.name}._derived_process_mqtt_message  path:0x{path_dbg:08X}",
        )



from gwproactor_test.certs import (
    TEST_CA_CERTIFICATE_PATH_VAR,
    TEST_CA_PRIVATE_KEY_VAR,
    TEST_CERTIFICATE_CACHE_VAR,
    copy_keys,
    set_test_certificate_cache_dir,
    test_ca_certificate_path,
    test_ca_private_key_path,
    test_certificate_cache_dir,
)
from gwproactor_test.clean import DefaultTestEnv, clean_test_env, default_test_env
from gwproactor_test.collections import ProactorCommTests
from gwproactor_test.comm_test_helper import CommTestHelper, ProactorTestHelper
from gwproactor_test.logger_guard import LoggerGuard, LoggerGuards, restore_loggers
from gwproactor_test.recorder import (
    ProactorT,
    RecorderInterface,
    RecorderLinkStats,
    RecorderStats,
    make_recorder_class,
)
from gwproactor_test.wait import (
    AwaitablePredicate,
    ErrorStringFunction,
    Predicate,
    StopWatch,
    await_for,
)

__all__ = [
    "TEST_CA_CERTIFICATE_PATH_VAR",
    "TEST_CA_PRIVATE_KEY_VAR",
    "TEST_CERTIFICATE_CACHE_VAR",
    "AwaitablePredicate",
    "CommTestHelper",
    "DefaultTestEnv",
    "ErrorStringFunction",
    "LoggerGuard",
    "LoggerGuards",
    "Predicate",
    "ProactorCommTests",
    "ProactorT",
    "ProactorTestHelper",
    "RecorderInterface",
    "RecorderLinkStats",
    "RecorderStats",
    "StopWatch",
    "await_for",
    "clean_test_env",
    "copy_keys",
    "default_test_env",
    "make_recorder_class",
    "restore_loggers",
    "set_test_certificate_cache_dir",
    "test_ca_certificate_path",
    "test_ca_private_key_path",
    "test_certificate_cache_dir",
]



# ruff: noqa: PLR2004, ERA001
import typing
from typing import Type

import pytest

from gwproactor.links import StateName
from gwproactor_test.dummies.tree.messages import RelayInfoReported
from gwproactor_test.recorder import RecorderLinkStats
from gwproactor_test.tree_comm_test_helper import TreeCommTestHelper
from gwproactor_test.wait import await_for


@pytest.mark.asyncio
class ProactorTreeCommBasicTests:
    CTH: Type[TreeCommTestHelper]

    async def test_tree_no_parent(self) -> None:
        async with self.CTH() as h:
            # add child 1
            h.add_child()
            child1 = h.child1
            stats1 = child1.stats
            stats1to2 = stats1.link(child1.downstream_client)
            stats1toAtn = stats1.link(child1.upstream_client)
            counts1to2 = stats1to2.comm_event_counts
            counts1toAtn = stats1toAtn.comm_event_counts
            link1to2 = child1.links.link(child1.downstream_client)
            link1toAtn = child1.links.link(child1.upstream_client)
            assert stats1.num_received == 0
            assert link1to2.state == StateName.not_started
            assert link1toAtn.state == StateName.not_started

            # start child 1
            h.start_child1()
            await await_for(
                lambda: link1to2.active_for_send() and link1toAtn.active_for_send(),
                1,
                "ERROR waiting child1 links to be active_for_send",
                err_str_f=h.summary_str,
            )
            assert link1to2.state == StateName.awaiting_peer
            assert link1toAtn.state == StateName.awaiting_peer
            assert counts1to2["gridworks.event.comm.mqtt.connect"] == 1
            assert counts1to2["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert len(stats1to2.comm_events) == 2
            assert counts1toAtn["gridworks.event.comm.mqtt.connect"] == 1
            assert counts1toAtn["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert len(stats1toAtn.comm_events) == 2

            # add child 2
            h.add_child2()
            child2 = h.child2
            stats2 = child2.stats
            stats2to1 = stats2.link(child2.upstream_client)
            counts2to1 = stats2to1.comm_event_counts
            link2to1 = child2.links.link(child2.upstream_client)
            assert stats2.num_received == 0
            assert link2to1.state == StateName.not_started

            # start child 2
            h.start_child2()
            await await_for(
                lambda: link1to2.active() and link2to1.active(),
                1,
                "ERROR waiting child2 links to be active",
                err_str_f=h.summary_str,
            )
            assert link1to2.state == StateName.active
            assert link2to1.state == StateName.active
            assert counts1to2["gridworks.event.comm.peer.active"] == 1
            assert len(stats1to2.comm_events) == 3
            assert counts2to1["gridworks.event.comm.peer.active"] == 1
            assert counts2to1["gridworks.event.comm.mqtt.connect"] == 1
            assert counts2to1["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert len(stats2to1.comm_events) == 3

    async def test_tree_message_exchange(self) -> None:
        async with self.CTH(
            start_child1=True,
            start_child2=True,
        ) as h:
            child1 = h.child
            stats1 = child1.stats.link(child1.downstream_client)
            link1to2 = child1.links.link(child1.downstream_client)
            link1toAtn = child1.links.link(child1.upstream_client)

            child2 = h.child2
            stats2 = child2.stats.link(child2.upstream_client)
            link2to1 = child2.links.link(child2.upstream_client)

            # Wait for children to connect
            await await_for(
                lambda: link1to2.active()
                and link2to1.active()
                and link1toAtn.active_for_send(),
                1,
                "ERROR waiting children to connect",
                err_str_f=h.summary_str,
            )

            # exchange messages
            assert stats2.num_received_by_type["gridworks.dummy.set.relay"] == 0
            relay_name = "scada2.relay1"
            child1.set_relay(relay_name, True)

            # wait for response to be received
            await await_for(
                lambda: stats1.num_received_by_type["gridworks.event.relay.report"]
                == 1,
                1,
                "ERROR waiting child1 to receive relay report from child2",
                err_str_f=h.summary_str,
            )
            assert stats2.num_received_by_type["gridworks.dummy.set.relay"] == 1
            assert child1.relays.Relays == {relay_name: RelayInfoReported(Closed=True)}
            assert child2.relays == {relay_name: True}
            assert child1.relays.TotalChangeMismatches == 0

    @pytest.mark.asyncio
    async def test_tree_parent_comm(self) -> None:
        async with self.CTH(add_child=True) as h:
            h.start_child1()
            await await_for(
                h.child1.mqtt_quiescent,
                3,
                "child1.mqtt_quiescent",
                err_str_f=h.summary_str,
            )
            h.add_child2()
            h.start_child2()
            link1to2 = h.child1.links.link(h.child1.downstream_client)
            link2to1 = h.child2.links.link(h.child2.upstream_client)
            await await_for(
                lambda: link1to2.active() and link2to1.active(),
                3,
                "link1to2.active() and link2to1.active()",
                err_str_f=h.summary_str,
            )
            h.add_parent()
            h.start_parent()
            link1toAtn = h.child1.links.link(h.child1.upstream_client)
            linkAtnto1 = h.parent.links.link(h.parent.downstream_client)
            await await_for(
                lambda: link1toAtn.active() and linkAtnto1.active(),
                3,
                "link1toAtn.active() and linkAtnto1.active()",
                err_str_f=h.summary_str,
            )

    @pytest.mark.asyncio
    async def test_tree_event_forward(self) -> None:
        # child1_settings = DummyScada1Settings(
        #     logging=LoggingSettings(
        #         levels=LoggerLevels(
        #             message_summary=logging.INFO,
        #         )
        #     ),
        #     mqtt_link_poll_seconds=5,
        # )
        async with self.CTH(
            # child_settings=child1_settings,
            start_child=True,
            start_child2=True,
            start_parent=True,
            # verbose=True,
            # child1_verbose=True,
            # child2_verbose=True,
            child2_on_screen=False,
            parent_on_screen=False,
        ) as h:
            link1to2 = h.child1.links.link(h.child1.downstream_client)
            link2to1 = h.child2.links.link(h.child2.upstream_client)
            link1toAtn = h.child1.links.link(h.child1.upstream_client)
            linkAtnto1 = h.parent.links.link(h.parent.downstream_client)
            await await_for(
                lambda: (
                    link1toAtn.active()
                    and linkAtnto1.active()
                    and link1to2.active()
                    and link2to1.active()
                ),
                3,
                "link1toAtn.active() and linkAtnto1.active()",
                err_str_f=h.summary_str,
            )
            relay_name = "scada2.relay1"
            h.child1.set_relay(relay_name, True)

            statsAtnTo1 = typing.cast(
                RecorderLinkStats, h.parent.stats.link(h.parent.downstream_client)
            )

            def _atn_heard_reports() -> bool:
                es1 = statsAtnTo1.event_counts.get(str(h.child1.publication_name))
                es2 = statsAtnTo1.event_counts.get(str(h.child2.publication_name))
                return bool(
                    es1
                    and es1["gridworks.event.relay.report.received"] == 1
                    and es2
                    and es2["gridworks.event.relay.report"] == 1
                )

            await await_for(
                _atn_heard_reports,
                1,
                "ERROR waiting for atn to hear reports",
                err_str_f=h.summary_str,
            )



import uuid
from typing import Literal

from gwproto import Message
from gwproto.messages import EventBase
from pydantic import BaseModel, Field


class RelayInfo(BaseModel):
    RelayName: str = ""
    Closed: bool = False


class RelayInfoReported(RelayInfo):
    CurrentChangeMismatch: bool = False
    MismatchCount: int = 0


class RelayStates(BaseModel):
    TotalChangeMismatches: int = 0
    Relays: dict[str, RelayInfoReported] = {}
    TypeName: Literal["gridworks.dummy.relay.states"] = "gridworks.dummy.relay.states"


class SetRelay(RelayInfo):
    MessageId: str = Field(default_factory=lambda: str(uuid.uuid4()))
    TypeName: Literal["gridworks.dummy.set.relay"] = "gridworks.dummy.set.relay"


class SetRelayMessage(Message[SetRelay]):
    def __init__(
        self,
        *,
        src: str,
        relay_name: str,
        closed: bool,
        dst: str = "",
        ack_required: bool = False,
    ) -> None:
        super().__init__(
            Src=src,
            Dst=dst,
            AckRequired=ack_required,
            Payload=SetRelay(RelayName=relay_name, Closed=closed),
        )


class RelayReportEvent(EventBase):
    """Dummy event, scada2 -> scada1"""

    relay_name: str = ""
    closed: bool = False
    changed: bool = False
    TypeName: Literal["gridworks.event.relay.report"] = "gridworks.event.relay.report"


class RelayReportReceivedEvent(RelayReportEvent):
    """Dummy event, scada1 *received* RelayReportEvent"""

    mismatch: bool = False
    mismatch_count: int = 0
    TypeName: Literal["gridworks.event.relay.report.received"] = (
        "gridworks.event.relay.report.received"
    )



import argparse
import logging
import textwrap
from pathlib import Path
from typing import Any, Callable, Optional, Self, Type

from gwproactor import Proactor, ProactorSettings, setup_logging
from gwproactor.config import DEFAULT_BASE_NAME, LoggingSettings, MQTTClient, Paths
from gwproactor_test.comm_test_helper import (
    ChildSettingsT,
    CommTestHelper,
    ProactorTestHelper,
)
from gwproactor_test.dummies import DUMMY_ATN_NAME, DUMMY_SCADA1_NAME, DUMMY_SCADA2_NAME
from gwproactor_test.logger_guard import LoggerGuards
from gwproactor_test.recorder import (
    ProactorT,
    RecorderInterface,
    make_recorder_class,
)


class TreeCommTestHelper(CommTestHelper):
    child2_t: Type[Proactor]
    child2_settings_t: Type[ProactorSettings]
    child2_recorder_t: Callable[..., RecorderInterface] = None
    child2_helper: ProactorTestHelper
    child2_verbose: bool = False
    child2_on_screen: bool = False
    child2_logger_guards: LoggerGuards

    def __init__(
        self,
        *,
        add_child1: bool = False,
        start_child1: bool = False,
        child1_verbose: bool = False,
        child2_settings: Optional[ChildSettingsT] = None,
        child2_verbose: bool = False,
        add_child2: bool = False,
        start_child2: bool = False,
        child2_on_screen: bool = False,
        child2_name: str = DUMMY_SCADA2_NAME,
        child_path_name: str = DUMMY_SCADA1_NAME,
        parent_path_name: str = DUMMY_ATN_NAME,
        child2_path_name: str = DUMMY_SCADA2_NAME,
        child2_kwargs: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        if not kwargs.get("child_name"):
            kwargs["child_name"] = DUMMY_SCADA1_NAME
        if not kwargs.get("parent_name"):
            kwargs["parent_name"] = DUMMY_ATN_NAME
        kwargs["add_child"] = add_child1 or kwargs.get("add_child", False)
        kwargs["start_child"] = start_child1 or kwargs.get("start_child", False)
        kwargs["child_verbose"] = child1_verbose or kwargs.get("child_verbose", False)
        super().__init__(
            child_path_name=child_path_name,
            parent_path_name=parent_path_name,
            **kwargs,
        )
        self.setup_child2_class()
        self.child2_helper = ProactorTestHelper(
            name=child2_name,
            path_name=child2_path_name,
            settings=(
                self.child2_settings_t(
                    logging=LoggingSettings(
                        base_log_name=f"{child2_path_name}_{DEFAULT_BASE_NAME}"
                    ),
                    paths=Paths(name=Path(child2_path_name)),
                )
                if child2_settings is None
                else child2_settings
            ),
            kwargs={} if child2_kwargs is None else child2_kwargs,
        )
        self.child2_verbose = child2_verbose
        self.child2_on_screen = child2_on_screen
        self.setup_child2_logging()
        if add_child2 or start_child2:
            self.add_child2()
            if start_child2:
                self.start_child2()

    @classmethod
    def setup_child2_class(cls) -> None:
        if cls.child2_recorder_t is None:
            cls.child2_recorder_t = make_recorder_class(cls.child2_t)

    @property
    def child1(self) -> Optional[ProactorT]:
        return self.child

    def add_child1(self) -> Self:
        return self.add_child()

    def start_child1(self) -> "CommTestHelper":
        return self.start_child()

    def make_child2(self) -> RecorderInterface:
        return self._make(self.child2_recorder_t, self.child2_helper)

    @property
    def child2(self) -> Optional[ProactorT]:
        return self.child2_helper.proactor

    def start_child2(self) -> Self:
        if self.child2 is not None:
            self.start_proactor(self.child2)
        return self

    def add_child2(self) -> Self:
        self.child2_helper.proactor = self.make_child2()
        return self

    def remove_child2(self) -> Self:
        self.child2_helper.proactor = None
        return self

    def _get_child2_clients_supporting_tls(self) -> list[MQTTClient]:
        return self._get_clients_supporting_tls(self.child2_helper.settings)

    def set_use_tls(self, use_tls: bool) -> None:
        super().set_use_tls(use_tls)
        self._set_settings_use_tls(use_tls, self._get_child2_clients_supporting_tls())

    def setup_child2_logging(self) -> None:
        self.child2_helper.settings.paths.mkdirs(parents=True)
        errors = []
        if not self.lifecycle_logging and not self.verbose and not self.child2_verbose:
            self.child2_helper.settings.logging.levels.lifecycle = logging.WARNING
        self.logger_guards.add_loggers(
            list(self.child2_helper.settings.logging.qualified_logger_names().values())
        )
        setup_logging(
            argparse.Namespace(verbose=self.verbose or self.child2_verbose),
            self.child2_helper.settings,
            errors=errors,
            add_screen_handler=self.child2_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    def get_proactors(self) -> list[RecorderInterface]:
        proactors = super().get_proactors()
        if self.child2_helper.proactor is not None:
            proactors.append(self.child2_helper.proactor)
        return proactors

    def get_log_path_str(self, exc: BaseException) -> str:
        return (
            f"CommTestHelper caught error {exc}.\n"
            "Working log dirs:"
            f"\n\t[{self.child_helper.settings.paths.log_dir}]"
            f"\n\t[{self.parent_helper.settings.paths.log_dir}]"
        )

    def summary_str(self) -> str:
        s = ""
        if self.child1 is None:
            s += "SCADA1: None\n"
        else:
            s += "SCADA1:\n"
            s += textwrap.indent(self.child1.summary_str(), "    ") + "\n"
        if self.child2 is None:
            s += "SCADA2: None\n"
        else:
            s += "SCADA2:\n"
            s += textwrap.indent(self.child2.summary_str(), "    ") + "\n"
        if self.parent is None:
            s += "ATN: None\n"
        else:
            s += "ATN:\n"
            s += textwrap.indent(self.parent.summary_str(), "    ") + "\n"
        return s



import platform
import ssl
import sys
import uuid
from enum import StrEnum, auto
from typing import Any

import rich
from gwproto import Message, MQTTTopic
from gwproto.messages import Ack
from paho.mqtt.client import Client as PahoMQTTClient
from paho.mqtt.client import MQTTMessage

from gwproactor_test.dummies.names import DUMMY_ADMIN_NAME
from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandReadRelays,
    AdminCommandSetRelay,
    AdminInfo,
)
from gwproactor_test.dummies.tree.admin_settings import (
    AdminLinkSettings,
    DummyAdminSettings,
)
from gwproactor_test.dummies.tree.messages import RelayInfo, RelayStates


class AppState(StrEnum):
    not_started = auto()
    awaiting_connect = auto()
    awaiting_suback = auto()
    awaiting_command_ack = auto()
    awaiting_report = auto()
    stopped = auto()


class MQTTAdmin:
    client: PahoMQTTClient
    settings: DummyAdminSettings
    relay_name: str
    closed: bool
    user: str
    json: bool
    command_message_id: str
    state: AppState

    @property
    def mqtt_config(self) -> AdminLinkSettings:
        return self.settings.link

    def __init__(
        self,
        *,
        settings: DummyAdminSettings,
        relay_name: str,
        closed: bool,
        user: str,
        json: bool,
    ) -> None:
        self.settings = settings
        self.relay_name = relay_name
        self.closed = closed
        self.user = user
        self.json = json
        self.command_message_id = ""
        self.state = AppState.not_started
        self.client = PahoMQTTClient("-".join(str(uuid.uuid4()).split("-")[:-1]))
        self.client.username_pw_set(
            username=self.mqtt_config.username,
            password=self.mqtt_config.password.get_secret_value(),
        )
        tls_config = self.mqtt_config.tls
        if tls_config.use_tls:
            self.client.tls_set(
                ca_certs=tls_config.paths.ca_cert_path,
                certfile=tls_config.paths.cert_path,
                keyfile=tls_config.paths.private_key_path,
                cert_reqs=tls_config.cert_reqs,
                tls_version=ssl.PROTOCOL_TLS_CLIENT,
                ciphers=tls_config.ciphers,
                keyfile_password=tls_config.keyfile_password.get_secret_value(),
            )
        self.client.on_message = self.on_message
        self.client.on_connect = self.on_connect
        self.client.on_connect_fail = self.on_connect_fail
        self.client.on_disconnect = self.on_disconnect
        self.client.on_subscribe = self.on_subscribe

    def run(self) -> None:
        if not self.json:
            rich.print(f"Connecting to broker at <{self.mqtt_config.host}>")
        self.state = AppState.awaiting_connect
        self.client.connect(
            self.mqtt_config.host, port=self.mqtt_config.effective_port()
        )
        self.client.loop_forever()

    def on_connect(self, _: Any, _userdata: Any, _flags: dict, _rc: int) -> None:
        topic = MQTTTopic.encode(
            envelope_type=Message.type_name(),
            src=self.settings.target_gnode,
            dst=self.mqtt_config.long_name,
            message_type="#",
        )
        self.state = AppState.awaiting_suback
        if not self.json:
            rich.print(f"Connected. Subscribing to <{topic}>")
        self.client.subscribe(topic=topic)

    def on_subscribe(
        self, _: Any, _userdata: Any, _mid: int, _granted_qos: list[int]
    ) -> None:
        self.state = AppState.awaiting_command_ack
        message = Message[AdminCommandSetRelay](
            Src=DUMMY_ADMIN_NAME,
            Dst=self.settings.target_gnode,
            MessageId=str(uuid.uuid4()),
            AckRequired=True,
            Payload=AdminCommandSetRelay(
                CommandInfo=AdminInfo(
                    User=self.user,
                    SrcMachine=platform.node(),
                ),
                RelayInfo=RelayInfo(
                    RelayName=self.relay_name,
                    Closed=self.closed,
                ),
            ),
        )
        self.command_message_id = message.Payload.MessageId
        topic = message.mqtt_topic()
        if not self.json:
            rich.print("Subscribed. Sending:")
            rich.print(message)
            rich.print(f"to topic <{topic}>")
        self.client.publish(topic=topic, payload=message.model_dump_json().encode())

    def on_connect_fail(self, _: Any, _userdata: Any) -> None:
        if not self.json:
            rich.print("Connect failed. Exiting")
        self.state = AppState.stopped
        self.client.loop_stop()
        sys.exit(1)

    def on_disconnect(self, _: Any, _userdata: Any, _rc: int) -> None:
        if not self.json:
            rich.print("Disconnected. Exiting")
        self.state = AppState.stopped
        self.client.loop_stop()
        sys.exit(2)

    def on_message(self, _: Any, _userdata: Any, message: MQTTMessage) -> None:
        if not self.json:
            rich.print(f"Received <{message.topic}>  in state <{self.state}>")
        msg_type = message.topic.split("/")[-1].replace("-", ".")
        if (
            self.state == AppState.awaiting_command_ack
            and msg_type == Ack.model_fields["TypeName"].default
        ):
            ack_message = Message[Ack].model_validate_json(message.payload)
            if ack_message.Payload.AckMessageID == self.command_message_id:
                self.state = AppState.awaiting_report
                message = Message[AdminCommandReadRelays](
                    Src=self.mqtt_config.long_name,
                    Dst=self.settings.target_gnode,
                    MessageId=str(uuid.uuid4()),
                    AckRequired=True,
                    Payload=AdminCommandReadRelays(
                        CommandInfo=AdminInfo(
                            User=self.user,
                            SrcMachine=platform.node(),
                        ),
                    ),
                )
                self.command_message_id = message.Payload.MessageId
                topic = message.mqtt_topic()
                if not self.json:
                    rich.print("Subscribed. Sending:")
                    rich.print(message)
                    rich.print(f"at topic <{topic}>")
                self.client.publish(
                    topic=topic, payload=message.model_dump_json().encode()
                )
            else:
                if not self.json:
                    rich.print(
                        "Received unexpected ack for "
                        f"{ack_message.Payload.AckMessageID}. Expected: "
                        f"{self.command_message_id}. Exiting."
                    )
                self.state = AppState.stopped
                self.client.loop_stop()
                sys.exit(3)
        elif (
            self.state == AppState.awaiting_report
            and msg_type == RelayStates.model_fields["TypeName"].default
        ):
            report_message = Message[RelayStates].model_validate_json(message.payload)
            if self.json:
                print(report_message.Payload.model_dump_json(indent=2))  # noqa
            else:
                rich.print(report_message.Payload)
            self.state = AppState.stopped
            self.client.loop_stop()
            sys.exit(0)



# ruff: noqa: ERA001
import asyncio
import contextlib
import subprocess  # noqa: S404
import time
from typing import Optional

from gwproto import Message

from gwproactor.message import (
    InternalShutdownMessage,
    KnownNames,
    PatExternalWatchdog,
    PatExternalWatchdogMessage,
    PatInternalWatchdog,
)
from gwproactor.proactor_interface import (
    Communicator,
    MonitoredName,
    Runnable,
    ServicesInterface,
)


class _MonitoredName(MonitoredName):
    last_pat: float = 0.0


class WatchdogManager(Communicator, Runnable):
    _watchdog_task: Optional[asyncio.Task] = None
    _seconds_per_pat: float
    _monitored_names: dict[str, _MonitoredName]
    _pat_external_watchdog_process_args: list[str]

    def __init__(self, seconds_per_pat: float, services: ServicesInterface) -> None:
        super().__init__(KnownNames.watchdog_manager.value, services)
        self.lg = services.logger
        self._seconds_per_pat = seconds_per_pat
        self._monitored_names = {}
        self._pat_external_watchdog_process_args = []

    def start(self) -> None:
        if self._watchdog_task is None:
            self._pat_external_watchdog_process_args = (
                self._services.get_external_watchdog_builder_class().pat_args(
                    str(self._services.settings.paths.name),
                )
            )
            self.lg.info(
                f"WatchdogManager: [{' '.join(self._pat_external_watchdog_process_args)}]"
            )
            now = time.time()
            for monitored in self._monitored_names.values():
                monitored.last_pat = now
            self._watchdog_task = asyncio.create_task(
                self._check_pats(), name="pat_watchdog"
            )

    def stop(self) -> None:
        if self._watchdog_task is not None and not self._watchdog_task.done():
            self._watchdog_task.cancel()

    async def join(self) -> None:
        if self._watchdog_task is not None:
            with contextlib.suppress(asyncio.CancelledError):
                await self._watchdog_task

    def process_message(self, message: Message) -> None:
        # self.lg.path("++WatchdogManager.process_message")
        path_dbg = 0
        match message.Payload:
            case PatInternalWatchdog():
                path_dbg |= 0x00000001
                self._pat_internal_watchdog(message.src())
            case PatExternalWatchdog():
                path_dbg |= 0x00000002
                self._pat_external_watchdog()
            case _:
                path_dbg |= 0x00000004
                raise ValueError(
                    f"WatchdogManager does not handle message payloads of type {type(message.Payload)}"
                )
        # self.lg.path(f"--WatchdogManager.process_message  0x{path_dbg:08X}")

    def _pat_internal_watchdog(self, name: str) -> None:
        if name not in self._monitored_names:
            raise ValueError(
                f"ERROR. Received interal watchdog pat from unmonitored name: {name}. Monistored names: {list(self._monitored_names.keys())}"
            )
        self._monitored_names[name].last_pat = time.time()

    def add_monitored_name(self, monitored: MonitoredName) -> None:
        if monitored.timeout_seconds <= self._seconds_per_pat / 2:
            raise ValueError(
                f"ERROR. WatchdogManager cannot reliably monitor a timeout of {monitored.timeout_seconds} "
                f"(requested for {monitored.name}) because "
                f"WatchdogManager's _seconds_per_pat (sample rate) is {self._seconds_per_pat}"
            )
        if monitored.name in self._monitored_names:
            raise ValueError(
                f"ERROR. Name {monitored.name} is already being monitored with {self._monitored_names[monitored.name]}"
            )
        self._monitored_names[monitored.name] = _MonitoredName(
            monitored.name, monitored.timeout_seconds
        )

    def _check_for_timeout_expired(self) -> Optional[_MonitoredName]:
        # self.lg.path("++_check_for_timeout_expired")
        # path_dbg = 0
        expired: Optional[_MonitoredName] = None
        now = time.time()
        for monitored in self._monitored_names.values():
            # path_dbg |= 0x0000001
            required_pat_time = monitored.last_pat + monitored.timeout_seconds
            # self.lg.info(
            #     f"  {monitored.name:50s}  "
            #     f"{monitored.timeout_seconds:4d}  last_pat:{monitored.last_pat:11.1f}  "
            #     f"required_pat_time: {required_pat_time:11.1f}  "
            #     f"now:{now:11.1f}  "
            #     f"remaining: {int(required_pat_time - now):4d}  "
            #     f"required_pat_time < now ? {int(required_pat_time < now)}"
            # )
            if required_pat_time < now:
                # path_dbg |= 0x0000002
                expired = monitored
                break
        # self.lg.path(f"--_check_for_timeout_expired: count:{int(expired is not None)}  0x{path_dbg:08X}")
        return expired

    async def _check_pats(self) -> None:
        while (expired := self._check_for_timeout_expired()) is None:
            self._send(PatExternalWatchdogMessage())
            await asyncio.sleep(self._seconds_per_pat)
        self._send(
            InternalShutdownMessage(
                Src=self.name,
                Reason=(
                    f"Monitored object ({expired.name}) failed to pat internal watchdog.  \n"
                    f"  Last pat from {expired.name}: {int(time.time() - expired.last_pat)} seconds ago\n"
                    f"  Allowed seconds: {int(expired.timeout_seconds)}"
                ),
            )
        )

    def _pat_external_watchdog(self) -> None:
        if self._pat_external_watchdog_process_args:
            subprocess.run(self._pat_external_watchdog_process_args, check=True)  # noqa: S603



# ruff: noqa: PLR2004, ERA001

import warnings
from typing import Any, Type

import pytest
from gwproto import MQTTTopic
from paho.mqtt.client import MQTT_ERR_CONN_LOST

from gwproactor.links import StateName
from gwproactor.message import DBGPayload
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.wait import await_for


@pytest.mark.asyncio
class ProactorCommAwaitingSetupTests:
    CTH: Type[CommTestHelper]

    @pytest.mark.asyncio
    async def test_awaiting_setup_and_peer(self) -> None:
        """
        Test:
         (connecting -> connected -> awaiting_setup_and_peer)
         (awaiting_setup_and_peer -> mqtt_suback -> awaiting_peer)
         (awaiting_setup_and_peer -> disconnected -> connecting)
        """
        async with self.CTH(add_child=True) as h:
            child = h.child
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child.links.link(child.upstream_client)

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started

            # start child
            child.pause_upstream_subacks()
            h.start_child()
            await await_for(
                lambda: child.num_upstream_subacks_available() == 1,
                1,
                "ERROR waiting suback pending",
                err_str_f=h.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Allow suback to arrive
            child.release_upstream_subacks()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                1,
                "ERROR waiting mqtt_suback",
                err_str_f=h.summary_str,
            )
            assert link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 2
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Tell client we lost comm
            child.pause_upstream_subacks()
            child.mqtt_client_wrapper(  # noqa: SLF001
                child.upstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            await await_for(
                lambda: child.num_upstream_subacks_available() == 1,
                3,
                "ERROR waiting suback pending",
                err_str_f=h.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 4
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Tell client we lost comm
            child.clear_upstream_subacks()
            child.mqtt_client_wrapper(  # noqa: SLF001
                child.upstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            await await_for(
                lambda: len(stats.comm_events) > 4,
                1,
                "ERROR waiting comm fail",
                err_str_f=h.summary_str,
            )
            await await_for(
                lambda: link.in_state(StateName.awaiting_setup_and_peer),
                3,
                "ERROR waiting comm restore",
                err_str_f=h.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 3
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 2
            assert len(stats.comm_events) == 6
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Allow suback to arrive
            child.release_upstream_subacks()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                1,
                "ERROR waiting mqtt_suback",
                err_str_f=h.summary_str,
            )
            assert link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 3
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 2
            assert len(stats.comm_events) == 7
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

    @pytest.mark.asyncio
    async def test_awaiting_setup_and_peer_corner_cases(self, request: Any) -> None:
        """
        Test corner cases:
         (connecting -> connected -> awaiting_setup_and_peer)
         (awaiting_setup_and_peer -> mqtt_suback -> awaiting_setup_and_peer)
         (awaiting_setup_and_peer -> mqtt_suback -> awaiting_peer)
         (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
        Force 1 suback per subscription. By default MQTTClientWrapper packs as many subscriptions as possible into a
        single subscribe message, so by default child only receives a single suback for all subscriptions.
        So that we can test (awaiting_setup_and_peer -> mqtt_suback -> awaiting_setup_and_peer) self-loop transition,
        which might occur if we have too many subscriptions for that to be possible, we force the suback response to
        be split into multiple messages.

        In practice these might be corner cases that rarely or never occur, since by default all subacks will come and
        one message and we should not receive any messages before subscribing.
        """
        async with self.CTH(add_child=True) as h:
            child = h.child
            child_subscriptions = child.mqtt_subscriptions(child.upstream_client)
            if len(child_subscriptions) < 2:
                if h.warn_if_multi_subscription_tests_skipped:
                    warnings.warn(
                        (
                            f"Skipping <{request.node.name}> because configured child proactor <{child.name}> "
                            f"has < 2 subscriptions. Subscriptions: {child_subscriptions}"
                        ),
                        stacklevel=2,
                    )
                return
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child.links.link(child.upstream_client)

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started

            # start child
            child.split_client_subacks(child.upstream_client)
            child.pause_upstream_subacks()
            h.start_child()
            await await_for(
                lambda: child.num_upstream_subacks_available() == 3,
                3,
                "ERROR waiting link reconnect",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Allow one suback at a time to arrive
            # suback 1/3
            # (mqtt_suback -> awaiting_setup_and_peer)
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_upstream_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # suback 2/3
            # (mqtt_suback -> awaiting_setup_and_peer)
            child.release_upstream_subacks(1)
            exp_subacks += 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (2/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # suback 3/3
            # (mqtt_suback -> awaiting_peer)
            child.release_upstream_subacks(1)
            exp_subacks += 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (3/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_peer
            assert not link.active_for_recv()
            assert not link.active()
            assert link.active_for_send()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 2
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # (message_from_peer -> awaiting_setup)
            # Tell client we lost comm
            child.pause_upstream_subacks()
            child.mqtt_client_wrapper(  # noqa: SLF001
                child.upstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            await await_for(
                lambda: child.num_upstream_subacks_available() == 3,
                3,
                "ERROR waiting suback pending",
                err_str_f=h.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 4
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Allow one suback at a time to arrive
            # (Not strictly necessary, since message receiving code does not check if the source topic suback
            #  has arrived).
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_upstream_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # Start the parent, wait for it to send us a message, which will
            # transition us into awaiting_setup
            h.add_parent()
            h.start_parent()
            await await_for(
                lambda: link.in_state(StateName.awaiting_setup),
                3,
                "ERROR waiting suback pending",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 4

    @pytest.mark.asyncio
    async def test_awaiting_setup2__(self, request: Any) -> None:
        """
        Test awaiting_setup (corner state):
         (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
         (awaiting_setup -> mqtt_suback -> awaiting_setup)
         (awaiting_setup -> mqtt_suback -> active)
         (awaiting_setup -> message_from_peer -> awaiting_setup)
         (awaiting_setup -> disconnected -> connecting)
        Force 1 suback per subscription. By default MQTTClientWrapper packs as many subscriptions as possible into a
        single subscribe message, so by default child only receives a single suback for all subscriptions.
        So that we can test (awaiting_setup_and_peer -> mqtt_suback -> awaiting_setup_and_peer) self-loop transition,
        which might occur if we have too many subscriptions for that to be possible, we force the suback response to
        be split into multiple messages.

        In practice these might be corner cases that rarely or never occur, since by default all subacks will come and
        one message and we should not receive any messages before subscribing.
        """
        async with self.CTH(add_child=True, add_parent=True) as h:
            child = h.child
            child_subscriptions = child.mqtt_subscriptions(child.upstream_client)
            if len(child_subscriptions) < 2:
                if h.warn_if_multi_subscription_tests_skipped:
                    warnings.warn(
                        (
                            f"Skipping <{request.node.name}> because configured child proactor <{child.name}> "
                            f"has < 2 subscriptions. Subscriptions: {child_subscriptions}"
                        ),
                        stacklevel=2,
                    )
                return
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child.links.link(child.upstream_client)

            parent = h.parent

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started

            # start child
            # (not_started -> started -> connecting)
            # (connecting -> connected -> awaiting_setup_and_peer)
            child.split_client_subacks(child.upstream_client)
            child.pause_upstream_subacks()
            h.start_child()
            await await_for(
                lambda: child.num_upstream_subacks_available() == 3,
                3,
                "ERROR waiting link reconnect",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Allow one suback at a time to arrive
            # (Not strictly necessary, since message receiving code does not check if the source topic suback
            #  has arrived).
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_upstream_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
            # Start the parent, wait for it to send us a message, which will
            # transition us into awaiting_setup
            h.start_parent()

            await await_for(
                lambda: link.in_state(StateName.awaiting_setup),
                3,
                "ERROR waiting suback pending",
                err_str_f=h.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1

            # (awaiting_setup -> mqtt_suback -> awaiting_setup)
            # Allow another suback to arrive, remaining in awaiting_setup
            child.release_upstream_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (2/3)",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup

            # (awaiting_setup -> message_from_peer -> awaiting_setup)
            # Receive another message from peer, remaining in awaiting_setup

            # noinspection PyTypeChecker
            dbg_topic = MQTTTopic.encode(
                "gw",
                parent.publication_name,
                parent.links.topic_dst(parent.downstream_client),
                DBGPayload.model_fields["TypeName"].default,
            )
            assert stats.num_received_by_topic[dbg_topic] == 0
            parent.send_dbg(parent.downstream_client)
            await await_for(
                lambda: stats.num_received_by_topic[dbg_topic] == 1,
                1,
                "ERROR waiting for dbg message",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup

            # (awaiting_setup -> disconnected -> connecting)
            # Tell client we lost comm
            child.clear_upstream_subacks()
            child.pause_upstream_subacks()
            child.mqtt_client_wrapper(  # noqa: SLF001
                child.upstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            await await_for(
                lambda: child.num_upstream_subacks_available() == 3,
                3,
                "ERROR waiting suback pending",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 3
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Allow one suback at a time to arrive
            # (Not strictly necessary, since message receiving code does not check if the source topic suback
            #  has arrived).
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_upstream_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=h.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
            # Force parent to restore comm, delivering a message, sending us to awaiting_setup
            parent.mqtt_client_wrapper(  # noqa: SLF001
                parent.downstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            await await_for(
                lambda: link.in_state(StateName.awaiting_setup),
                3,
                "ERROR waiting for message from peer",
                err_str_f=h.summary_str,
            )

            # (awaiting_setup -> mqtt_suback -> active)
            # Release all subacks, allowing child to go active
            child.release_upstream_subacks()
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for active",
                err_str_f=h.summary_str,
            )



from gwproactor.links.acks import (
    DEFAULT_ACK_DELAY,
    AckManager,
    AckTimerCallback,
    AckWaitInfo,
)
from gwproactor.links.asyncio_timer_manager import AsyncioTimerManager
from gwproactor.links.link_manager import LinkManager, LinkManagerTransition
from gwproactor.links.link_state import (
    CommLinkAlreadyExists,
    CommLinkMissing,
    InvalidCommStateInput,
    LinkState,
    LinkStates,
    RuntimeLinkStateError,
    StateName,
    Transition,
    TransitionName,
)
from gwproactor.links.message_times import LinkMessageTimes, MessageTimes
from gwproactor.links.mqtt import QOS, MQTTClients, MQTTClientWrapper, Subscription
from gwproactor.links.reuploads import Reuploads
from gwproactor.links.timer_interface import TimerManagerInterface

__all__ = [
    "DEFAULT_ACK_DELAY",
    "QOS",
    "AckManager",
    "AckTimerCallback",
    "AckWaitInfo",
    "AsyncioTimerManager",
    "CommLinkAlreadyExists",
    "CommLinkMissing",
    "InvalidCommStateInput",
    "LinkManager",
    "LinkManagerTransition",
    "LinkMessageTimes",
    "LinkState",
    "LinkStates",
    "MQTTClientWrapper",
    "MQTTClients",
    "MessageTimes",
    "Reuploads",
    "RuntimeLinkStateError",
    "StateName",
    "Subscription",
    "TimerManagerInterface",
    "Transition",
    "TransitionName",
]



from pydantic import model_validator
from pydantic_settings import SettingsConfigDict

from gwproactor import ProactorSettings
from gwproactor.config import MQTTClient
from gwproactor_test.dummies.names import DUMMY_PARENT_ENV_PREFIX, DUMMY_PARENT_NAME


class DummyParentSettings(ProactorSettings):
    child_mqtt: MQTTClient = MQTTClient()

    model_config = SettingsConfigDict(env_prefix=DUMMY_PARENT_ENV_PREFIX)

    @model_validator(mode="before")
    @classmethod
    def pre_root_validator(cls, values: dict) -> dict:
        return ProactorSettings.update_paths_name(values, DUMMY_PARENT_NAME)



# ruff: noqa: PLR2004, ERA001

from typing import Type

import pytest

from gwproactor_test.awaiting_setup_tests import ProactorCommAwaitingSetupTests
from gwproactor_test.basic_tests import ProactorCommBasicTests
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.reupload_tests import ProactorReuploadTests
from gwproactor_test.timeout_tests import ProactorCommTimeoutTests
from gwproactor_test.tree_comm_test_helper import TreeCommTestHelper
from gwproactor_test.tree_tests import ProactorTreeCommBasicTests


@pytest.mark.asyncio
class ProactorCommTests(
    ProactorCommBasicTests,
    ProactorCommAwaitingSetupTests,
    ProactorCommTimeoutTests,
    ProactorReuploadTests,
):
    CTH: Type[CommTestHelper]


class ProactorTreeCommTests(ProactorTreeCommBasicTests):
    CTH: Type[TreeCommTestHelper]



import argparse
import logging
import sys
import traceback
from pathlib import Path
from typing import Optional, Type, TypeVar

import dotenv
import rich
from pydantic import BaseModel
from pydantic_settings import BaseSettings

from gwproactor import Proactor, ProactorSettings, setup_logging
from gwproactor.config import MQTTClient
from gwproactor.config.paths import TLSPaths

LOGGING_FORMAT = "%(asctime)s %(message)s"


def missing_tls_paths(paths: TLSPaths) -> list[tuple[str, Optional[Path]]]:
    missing = []
    for path_name in paths.model_fields:
        path = getattr(paths, path_name)
        if path is None or not Path(path).exists():
            missing.append((path_name, path))
    return missing


def check_tls_paths_present(
    model: BaseModel | BaseSettings, *, raise_error: bool = True
) -> str:
    missing_str = ""
    for k in model.model_fields:
        v = getattr(model, k)
        if isinstance(v, MQTTClient) and v.tls.use_tls:
            missing_paths = missing_tls_paths(v.tls.paths)
            if missing_paths:
                missing_str += f"client {k}\n"
                for path_name, path in missing_paths:
                    missing_str += f"  {path_name:20s}  {path}\n"
    if missing_str:
        error_str = f"ERROR. TLS usage requested but the following files are missing:\n{missing_str}"
        if raise_error:
            raise ValueError(error_str)
    else:
        error_str = ""
    return error_str


ProactorT = TypeVar("ProactorT", bound=Proactor)
ProactorSettingsT = TypeVar("ProactorSettingsT", bound=ProactorSettings)


def get_settings(
    *,
    settings_type: Optional[Type[ProactorSettingsT]] = None,
    settings: Optional[ProactorSettingsT] = None,
    env_file: str | Path = ".env",
) -> ProactorSettingsT:
    if (settings_type is None and settings is None) or (
        settings_type is not None and settings is not None
    ):
        raise ValueError("ERROR. Specify exactly one of (settings_type, settings)")
    if settings_type is not None:
        settings = settings_type(_env_file=str(env_file))
    return settings


def print_settings(
    *,
    settings_type: Optional[Type[ProactorSettingsT]] = None,
    settings: Optional[ProactorSettingsT] = None,
    env_file: str | Path = ".env",
) -> None:
    dotenv_file = dotenv.find_dotenv(env_file)
    rich.print(
        f"Env file: <{dotenv_file}>  exists:{env_file and Path(dotenv_file).exists()}"
    )
    settings = get_settings(
        settings_type=settings_type, settings=settings, env_file=dotenv_file
    )
    rich.print(settings)
    missing_tls_paths_ = check_tls_paths_present(settings, raise_error=False)
    if missing_tls_paths_:
        rich.print(missing_tls_paths_)


def get_proactor(  # noqa: PLR0913
    name: str,
    proactor_type: Type[ProactorT],
    *,
    settings_type: Optional[Type[ProactorSettingsT]] = None,
    settings: Optional[ProactorSettingsT] = None,
    dry_run: bool = False,
    verbose: bool = False,
    message_summary: bool = False,
    env_file: str | Path = ".env",
    run_in_thread: bool = False,
    add_screen_handler: bool = True,
) -> ProactorT:
    dotenv_file = dotenv.find_dotenv(env_file)
    dotenv_file_debug_str = (
        f"Env file: <{dotenv_file}>  exists:{Path(dotenv_file).exists()}"
    )
    settings = get_settings(
        settings_type=settings_type, settings=settings, env_file=dotenv_file
    )
    if dry_run:
        rich.print(dotenv_file_debug_str)
        rich.print(settings)
        missing_tls_paths_ = check_tls_paths_present(settings, raise_error=False)
        if missing_tls_paths_:
            rich.print(missing_tls_paths_)
        rich.print("Dry run. Doing nothing.")
        sys.exit(0)
    else:
        settings.paths.mkdirs()
        args = argparse.Namespace(
            verbose=verbose,
            message_summary=message_summary,
        )
        setup_logging(args, settings, add_screen_handler=add_screen_handler)
        logger = logging.getLogger(
            settings.logging.qualified_logger_names()["lifecycle"]
        )
        logger.info("")
        logger.info(dotenv_file_debug_str)
        logger.info("Settings:")
        logger.info(settings.model_dump_json(indent=2))
        rich.print(settings)
        check_tls_paths_present(settings)
        proactor = proactor_type(name=name, settings=settings)
        if run_in_thread:
            logger.info("run_async_actors_main() starting")
            proactor.run_in_thread()
    return proactor


async def run_async_main(  # noqa: PLR0913
    name: str,
    proactor_type: Type[ProactorT],
    *,
    settings_type: Optional[Type[ProactorSettingsT]] = None,
    settings: Optional[ProactorSettingsT] = None,
    env_file: str | Path = ".env",
    dry_run: bool = False,
    verbose: bool = False,
    message_summary: bool = False,
) -> None:
    settings = get_settings(
        settings_type=settings_type,
        settings=settings,
        env_file=dotenv.find_dotenv(env_file),
    )
    exception_logger = logging.getLogger(settings.logging.base_log_name)
    try:
        proactor = get_proactor(
            name=name,
            proactor_type=proactor_type,
            settings=settings,
            dry_run=dry_run,
            verbose=verbose,
            message_summary=message_summary,
        )
        exception_logger = proactor.logger
        try:
            await proactor.run_forever()
        finally:
            proactor.stop()
    except SystemExit:
        pass
    except KeyboardInterrupt:
        pass
    except BaseException as e:
        try:
            exception_logger.exception(
                "ERROR in run_async_actors_main. Shutting down: " "[%s] / [%s]",
                e,  # noqa: TRY401
                type(e),  # noqa: TRY401
            )
        except:  # noqa: E722
            traceback.print_exception(e)
        raise



"""Actor: A partial implementation of ActorInterface which supplies the trivial implementations.

SyncThreadActor: An actor which orchestrates starting, stopping and communicating with a passed in
SyncAsyncInteractionThread
"""

from abc import ABC
from typing import Any, Generic, Sequence, TypeVar

from gwproto import Message
from gwproto.data_classes.sh_node import ShNode
from result import Result

from gwproactor.proactor_interface import (
    ActorInterface,
    Communicator,
    MonitoredName,
    ServicesInterface,
)
from gwproactor.sync_thread import SyncAsyncInteractionThread


class Actor(ActorInterface, Communicator, ABC):
    def __init__(self, name: str, services: ServicesInterface) -> None:
        self._node = services.hardware_layout.node(name)
        super().__init__(name, services)

    @property
    def name(self) -> str:
        return self._name

    @property
    # note this is over-written in the derived class ScadaActor
    # so that it is not static.
    def node(self) -> ShNode:
        return self._node

    def init(self) -> None:
        """Called after constructor so derived functions can be used in setup."""


SyncThreadT = TypeVar("SyncThreadT", bound=SyncAsyncInteractionThread)


class SyncThreadActor(Actor, Generic[SyncThreadT]):
    _sync_thread: SyncAsyncInteractionThread

    def __init__(
        self,
        name: str,
        services: ServicesInterface,
        sync_thread: SyncAsyncInteractionThread,
    ) -> None:
        super().__init__(name, services)
        self._sync_thread = sync_thread

    def process_message(self, message: Message) -> Result[bool, Exception]:
        raise ValueError(
            f"Error. {self.__class__.__name__} does not process any messages. Received {message.Header}"
        )

    def send_driver_message(self, message: Any) -> None:
        self._sync_thread.put_to_sync_queue(message)

    def start(self) -> None:
        self._sync_thread.set_async_loop_and_start(
            self.services.event_loop, self.services.async_receive_queue
        )

    def stop(self) -> None:
        self._sync_thread.request_stop()

    async def join(self) -> None:
        await self._sync_thread.async_join()

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        monitored_names = []
        if self._sync_thread.pat_timeout is not None:
            monitored_names.append(
                MonitoredName(self.name, self._sync_thread.pat_timeout)
            )
        return monitored_names



from gwproactor_test.dummies.pair.child import DummyChild
from gwproactor_test.dummies.pair.child_config import DummyChildSettings
from gwproactor_test.dummies.pair.parent import DummyParent
from gwproactor_test.dummies.pair.parent_config import DummyParentSettings

__all__ = [
    "DummyChild",
    "DummyChildSettings",
    "DummyParent",
    "DummyParentSettings",
]



# ruff: noqa: ERA001

import time
from dataclasses import dataclass
from pathlib import Path
from typing import Type

import pytest
from gwproto import MQTTTopic

from gwproactor import Proactor
from gwproactor.links import StateName
from gwproactor.message import DBGEvent, DBGPayload
from gwproactor.persister import TimedRollingFilePersister
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.wait import await_for


@dataclass
class _EventEntry:
    uid: str
    path: Path


class _EventGen:
    ok: list[_EventEntry]
    corrupt: list[_EventEntry]
    empty: list[_EventEntry]
    missing: list[_EventEntry]

    persister: TimedRollingFilePersister

    def __len__(self) -> int:
        return len(self.ok) + len(self.corrupt) + len(self.empty)

    def __init__(self, proactor: Proactor) -> None:
        self.ok = []
        self.corrupt = []
        self.empty = []
        self.missing = []
        persister = proactor.event_persister
        assert isinstance(persister, TimedRollingFilePersister)
        self.persister = persister

    def _generate_event(self, member_name: str) -> _EventEntry:
        event = DBGEvent(Command=DBGPayload(), Msg=f"event {len(self)} {member_name}")
        ret = self.persister.persist(
            event.MessageId, event.model_dump_json(indent=2).encode()
        )
        if ret.is_err():
            raise ret.err()
        entry = _EventEntry(
            event.MessageId,
            self.persister.get_path(event.MessageId),
        )
        getattr(self, member_name).append(entry)
        return entry

    def _generate_ok(self) -> _EventEntry:
        return self._generate_event("ok")

    def _generate_corrupt(self) -> _EventEntry:
        entry = self._generate_event("corrupt")
        with entry.path.open() as f:
            contents = f.read()
        with entry.path.open("w") as f:
            f.write(contents[:-6])
        return entry

    def _generate_empty(self) -> _EventEntry:
        entry = self._generate_event("empty")
        with entry.path.open("w") as f:
            f.write("")
        return entry

    def _generate_missing(self) -> _EventEntry:
        entry = self._generate_event("missing")
        entry.path.unlink()
        return entry

    def generate(
        self,
        num_ok: int = 0,
        num_corrupt: int = 0,
        num_empty: int = 0,
        num_missing: int = 0,
    ) -> None:
        for _ in range(num_ok):
            self._generate_ok()
        for _ in range(num_corrupt):
            self._generate_corrupt()
        for _ in range(num_empty):
            self._generate_empty()
        for _ in range(num_missing):
            self._generate_missing()


@pytest.mark.asyncio
class ProactorReuploadTests:
    CTH: Type[CommTestHelper]

    @pytest.mark.asyncio
    async def test_reupload_basic(self) -> None:
        """
        Test:
            reupload not requiring flow control
        """
        async with self.CTH(
            start_child=True,
            add_parent=True,
            verbose=False,
        ) as h:
            child = h.child
            child.disable_derived_events()
            upstream_link = h.child.links.link(child.upstream_client)
            reupload_counts = h.child.stats.link(child.upstream_client).reupload_counts
            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=h.summary_str,
            )
            # Some events should have been generated, and they should have all been sent
            assert child.links.num_pending > 0
            assert child.links.num_reupload_pending == 0
            assert child.links.num_reuploaded_unacked == 0
            assert not child.links.reuploading()
            assert reupload_counts.started == 0
            assert reupload_counts.completed == 0

            # Start parent, wait for reconnect.
            h.start_parent()
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for parent",
                err_str_f=h.summary_str,
            )

            # Wait for reuploading to complete
            await await_for(
                lambda: reupload_counts.completed > 0,
                1,
                "ERROR waiting for re-upload to complete",
                err_str_f=h.summary_str,
            )

            # All events should have been reuploaded.
            assert child.links.num_reupload_pending == 0
            assert child.links.num_reuploaded_unacked == 0
            assert not child.links.reuploading()

    @pytest.mark.asyncio
    async def test_reupload_flow_control_simple(self) -> None:
        """
        Test:
            reupload requiring flow control
        """
        async with self.CTH(
            start_child=True,
            add_parent=True,
            child_settings=self.CTH.child_settings_t(num_initial_event_reuploads=5),
            verbose=False,
        ) as h:
            child = h.child
            child.disable_derived_events()
            upstream_link = h.child.links.link(child.upstream_client)
            reupload_counts = h.child.stats.link(child.upstream_client).reupload_counts
            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=h.summary_str,
            )
            # Some events should have been generated, and they should have all been sent
            base_num_pending = child.links.num_pending
            assert base_num_pending > 0
            assert child.links.num_reupload_pending == 0
            assert child.links.num_reuploaded_unacked == 0
            assert not child.links.reuploading()

            # Generate more events than fit in pipe.
            events_to_generate = child.settings.num_initial_event_reuploads * 2
            for i in range(events_to_generate):
                child.generate_event(
                    DBGEvent(
                        Command=DBGPayload(),
                        Msg=f"event {i + 1} / {events_to_generate}",
                    )
                )
            child.logger.info(
                f"Generated {events_to_generate} events. Total pending events: {child.links.num_pending}"
            )

            # Start parent, wait for connect.
            h.start_parent()
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for parent",
                err_str_f=h.summary_str,
            )

            # Wait for reupload to complete
            await await_for(
                lambda: reupload_counts.completed > 0,
                1,
                "ERROR waiting for reupload to complete",
                err_str_f=h.summary_str,
            )

    @pytest.mark.asyncio
    async def test_reupload_flow_control_detail(self) -> None:
        """
        Test:
            reupload requiring flow control
        """
        async with self.CTH(
            start_child=True,
            add_parent=True,
            child_settings=self.CTH.child_settings_t(num_initial_event_reuploads=5),
            child_verbose=False,
            verbose=False,
            # parent_on_screen=True,
        ) as h:
            child = h.child
            child.disable_derived_events()
            child_links = h.child.links
            upstream_link = child_links.link(child.upstream_client)
            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=h.summary_str,
            )
            # Some events should happened already, through the startup and mqtt connect process, and they should have
            # all been sent.
            # These events include: There are at least 3 non-generated events: startup, (mqtt connect, mqtt subscribed)/mqtt client.
            base_num_pending = child_links.num_pending
            assert base_num_pending > 0
            assert child_links.num_reupload_pending == 0
            assert child_links.num_reuploaded_unacked == 0
            assert not child_links.reuploading()

            # Generate more events than fit in pipe.
            events_to_generate = child.settings.num_initial_event_reuploads * 2
            for i in range(events_to_generate):
                child.generate_event(
                    DBGEvent(
                        Command=DBGPayload(),
                        Msg=f"event {i + 1} / {events_to_generate}",
                    )
                )
            child.logger.info(
                f"Generated {events_to_generate} events. Total pending events: {child_links.num_pending}"
            )
            assert child_links.num_reupload_pending == 0
            assert child_links.num_reuploaded_unacked == 0
            assert not child_links.reuploading()

            # pause parent acks so that we watch flow control
            h.parent.pause_acks()

            # Start parent, wait for parent to be subscribed.
            h.start_parent()
            await await_for(
                lambda: h.parent.links.link_state(h.parent.downstream_client)
                == StateName.awaiting_peer,
                1,
                "ERROR waiting for parent awaiting_peer",
                err_str_f=h.summary_str,
            )

            # Wait for parent to have ping waiting to be sent
            await await_for(
                lambda: len(h.parent.links.needs_ack) > 0,
                1,
                "ERROR waiting for parent awaiting_peer",
                err_str_f=h.summary_str,
            )

            # release the ping
            h.parent.release_acks(num_to_release=1)

            # wait for child to receive ping
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for child peer_active",
                err_str_f=h.summary_str,
            )
            # There are 3 non-generated events: startup, mqtt connect, mqtt subscribed.
            # A "PeerActive" event is also pending but that is _not_ part of re-upload because it is
            # generated _after_ the peer is active (and therefore has its own ack timeout running, so does not need to
            # be managed by reupload).
            last_num_to_reupload = events_to_generate + base_num_pending
            last_num_reuploaded_unacked = child.settings.num_initial_event_reuploads
            last_num_repuload_pending = (
                last_num_to_reupload - child_links.num_reuploaded_unacked
            )
            err_s = (
                f"child_links.num_reuploaded_unacked: {child_links.num_reuploaded_unacked}\n"
                f"last_num_reuploaded_unacked:        {last_num_reuploaded_unacked}\n"
                f"child_links.num_reupload_pending:   {child_links.num_reupload_pending}\n"
                f"last_num_repuload_pending:          {last_num_repuload_pending}\n"
                f"{child.summary_str()}"
            )
            assert (
                child_links.num_reuploaded_unacked == last_num_reuploaded_unacked
            ), err_s
            assert child_links.num_reupload_pending == last_num_repuload_pending, err_s
            assert child_links.num_pending == last_num_to_reupload + 1
            assert child_links.reuploading()

            # noinspection PyTypeChecker
            parent_ack_topic = MQTTTopic.encode(
                "gw",
                h.parent.publication_name,
                h.child.subscription_name,
                "gridworks-ack",
            )
            acks_received_by_child = child.stats.num_received_by_topic[parent_ack_topic]

            # Release acks one by one.
            #
            #   Bound this loop by time, not by total number of acks since at least one non-reupload ack should arrive
            #   (for the PeerActive event) and others could arrive if, for example, a duplicate MQTT message appeared.
            #
            end_time = time.time() + 5
            # loop_count_dbg = 0
            acks_released = 0
            while child_links.reuploading() and time.time() < end_time:
                # loop_path_dbg = 0
                # loop_count_dbg += 1

                # release one ack
                acks_released += h.parent.release_acks(num_to_release=1)

                # Wait for child to receive an ack
                await await_for(
                    lambda: child.stats.num_received_by_topic[parent_ack_topic]
                    == acks_received_by_child + acks_released,  # noqa: B023
                    timeout=1,
                    tag=(
                        "ERROR waiting for child to receive ack "
                        f"(acks_released: {acks_released}) "
                        f"on topic <{parent_ack_topic}>"
                    ),
                    err_str_f=h.summary_str,
                )
                curr_num_reuploaded_unacked = child_links.num_reuploaded_unacked
                curr_num_repuload_pending = child_links.num_reupload_pending
                curr_num_to_reuplad = (
                    curr_num_reuploaded_unacked + curr_num_repuload_pending
                )
                if curr_num_to_reuplad == last_num_to_reupload:
                    # loop_path_dbg |= 0x00000001
                    assert curr_num_reuploaded_unacked == last_num_reuploaded_unacked
                    assert curr_num_repuload_pending == last_num_repuload_pending
                elif curr_num_to_reuplad == last_num_to_reupload - 1:
                    # loop_path_dbg |= 0x00000002
                    if curr_num_reuploaded_unacked == last_num_reuploaded_unacked:
                        assert (
                            curr_num_repuload_pending == last_num_repuload_pending - 1
                        )
                    else:
                        assert (
                            curr_num_reuploaded_unacked
                            == last_num_reuploaded_unacked - 1
                        )
                        assert curr_num_repuload_pending == last_num_repuload_pending
                    assert child_links.reuploading() == bool(
                        curr_num_reuploaded_unacked > 0
                    )
                else:
                    raise ValueError(
                        "Unexpected change in reupload counts: "
                        f"({last_num_reuploaded_unacked}, {last_num_repuload_pending}) -> "
                        f"({curr_num_reuploaded_unacked}, {curr_num_repuload_pending})"
                    )

                # child.logger.info(
                #     f"ack loop: {loop_count_dbg} / {acks_released}:"
                #     f"({last_num_reuploaded_unacked}, {last_num_repuload_pending}) -> "
                #     f"({curr_num_reuploaded_unacked}, {curr_num_repuload_pending})"
                #     f" loop_path_dbg: 0x{loop_path_dbg:08X}")

                last_num_to_reupload = curr_num_to_reuplad
                last_num_reuploaded_unacked = curr_num_reuploaded_unacked
                last_num_repuload_pending = curr_num_repuload_pending

            assert not child_links.reuploading()

    @pytest.mark.skip(
        reason="Test seems to gotten flakier; unclear if this is because test is too sensitive or because it is broken"
    )
    @pytest.mark.asyncio
    async def test_reupload_errors(self) -> None:
        async with self.CTH(
            start_child=True,
            add_parent=True,
            child_verbose=False,
        ) as h:
            child = h.child
            child.disable_derived_events()
            reupload_counts = h.child.stats.link(child.upstream_client).reupload_counts
            child_links = h.child.links
            upstream_link = child_links.link(child.upstream_client)
            parent = h.parent

            def _err_str() -> str:
                return (
                    f"\nCHILD\n{child.summary_str()}\n"
                    f"\nPARENT\n{parent.summary_str()}\n"
                )

            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=_err_str,
            )
            base_num_pending = child_links.num_pending
            assert base_num_pending > 0
            assert child_links.num_reupload_pending == 0
            assert child_links.num_reuploaded_unacked == 0
            assert not child_links.reuploading()

            generator = _EventGen(child)
            generator.generate(num_corrupt=10)
            generator.generate(num_ok=10)
            generator.generate(num_empty=10)
            generator.generate(num_ok=10)
            generator.generate(num_missing=10)
            generator.generate(num_ok=10)

            h.start_parent()
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for active",
                err_str_f=_err_str,
            )

            # Wait for reupload to complete
            await await_for(
                lambda: reupload_counts.completed > 0,
                3,
                "ERROR waiting for reupload to complete",
                err_str_f=_err_str,
            )
            assert reupload_counts.started == reupload_counts.completed
            assert parent.stats.num_events_received >= base_num_pending + 60



# ruff: noqa: PLR2004, ERA001

import asyncio
from typing import Type

import pytest
from gwproto import MQTTTopic

from gwproactor.links import StateName
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.wait import await_for


@pytest.mark.asyncio
class ProactorCommTimeoutTests:
    CTH: Type[CommTestHelper]

    @pytest.mark.asyncio
    async def test_response_timeout(self) -> None:
        """
        Test:
            (awaiting_peer -> response_timeout -> awaiting_peer)
            (active -> response_timeout -> awaiting_peer)
        """

        async with self.CTH(
            add_child=True,
            add_parent=True,
        ) as h:
            child = h.child
            link = child.links.link(child.upstream_client)
            stats = child.stats.link(child.upstream_client)
            parent = h.parent
            parent_link = parent.links.link(parent.downstream_client)

            # Timeout while awaiting setup
            # (awaiting_peer -> response_timeout -> awaiting_peer)

            # start parent
            parent.pause_acks()
            h.start_parent()
            await await_for(
                lambda: parent_link.in_state(StateName.awaiting_peer),
                3,
                "ERROR waiting for parent to connect to broker",
                err_str_f=parent.summary_str,
            )

            # start child
            child.set_ack_timeout_seconds(1)
            assert stats.timeouts == 0
            h.start_child()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                3,
                "ERROR waiting for child to connect to broker",
                err_str_f=parent.summary_str,
            )
            # (awaiting_peer -> response_timeout -> awaiting_peer)
            await await_for(
                lambda: stats.timeouts > 0,
                1,
                "ERROR waiting for child to timeout",
                err_str_f=parent.summary_str,
            )
            assert link.state == StateName.awaiting_peer
            assert child.event_persister.num_pending > 0

            # release the hounds
            # (awaiting_peer -> message_from_peer -> active)
            parent.release_acks()
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for parent to restore link #1",
                err_str_f=parent.summary_str,
            )
            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Timeout while active
            # (active -> response_timeout -> awaiting_peer)
            parent.pause_acks()
            child.force_ping(child.upstream_client)
            exp_timeouts = stats.timeouts + len(
                child.links.ack_manager._acks[child.upstream_client]  # noqa: SLF001
            )
            await await_for(
                lambda: stats.timeouts == exp_timeouts,
                1,
                "ERROR waiting for child to timeout",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_peer
            assert child.event_persister.num_pending > 0
            await await_for(
                lambda: len(parent.needs_ack) == 2,
                1,
                "ERROR waiting for child to timeout",
                err_str_f=child.summary_str,
            )

            # (awaiting_peer -> message_from_peer -> active)
            parent.release_acks()
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for parent to restore link #1",
                err_str_f=parent.summary_str,
            )

    @pytest.mark.skip(
        reason="Test seems to gotten flakier; unclear if this is because test is too sensitive or because it is broken"
    )
    @pytest.mark.asyncio
    async def test_ping(self) -> None:
        """
        Test:
            ping sent peridoically if no messages sent
            ping not sent if messages are sent
            ping restores comm
        """
        child_settings = self.CTH.child_settings_t()
        parent_settings = self.CTH.parent_settings_t()
        child_settings.mqtt_link_poll_seconds = (
            parent_settings.mqtt_link_poll_seconds
        ) = 0.1
        async with self.CTH(
            add_child=True,
            add_parent=True,
            child_settings=child_settings,
            parent_settings=parent_settings,
        ) as h:
            parent = h.parent
            parent_stats = parent.stats.link(parent.downstream_client)
            child = h.child
            # noinspection PyTypeChecker
            pings_from_parent_topic = MQTTTopic.encode(
                envelope_type="gw",
                src=parent.publication_name,
                dst=parent.links.topic_dst(parent.downstream_client),
                message_type="gridworks-ping",
            )
            child.disable_derived_events()
            child.set_ack_timeout_seconds(1)
            link = child.links.link(child.upstream_client)
            stats = child.stats.link(child.upstream_client)
            # noinspection PyTypeChecker
            ping_from_child_topic = MQTTTopic.encode(
                envelope_type="gw",
                src=child.publication_name,
                dst=child.links.topic_dst(child.downstream_client),
                message_type="gridworks-ping",
            )
            # start parent and child
            h.start_parent()
            h.start_child()
            await await_for(
                lambda: link.in_state(StateName.active),
                3,
                "ERROR waiting for child active",
                err_str_f=h.summary_str,
            )

            # Test that ping sent peridoically if no messages sent
            start_pings_from_parent = stats.num_received_by_topic[
                pings_from_parent_topic
            ]
            start_pings_from_child = parent_stats.num_received_by_topic[
                ping_from_child_topic
            ]
            start_messages_from_parent = stats.num_received
            start_messages_from_child = parent_stats.num_received
            wait_seconds = 0.5
            await asyncio.sleep(wait_seconds)
            pings_from_parent = (
                stats.num_received_by_topic[pings_from_parent_topic]
                - start_pings_from_parent
            )
            pings_from_child = (
                parent_stats.num_received_by_topic[ping_from_child_topic]
                - start_pings_from_child
            )
            messages_from_parent = stats.num_received - start_messages_from_parent
            messages_from_child = parent_stats.num_received - start_messages_from_child
            exp_pings_nominal = (
                wait_seconds / parent.settings.mqtt_link_poll_seconds
            ) - 1
            err_str = (
                f"\npings_from_parent: {pings_from_parent}  ({stats.num_received_by_topic[pings_from_parent_topic]} - {start_pings_from_parent})  on <{pings_from_parent_topic}>\n"
                f"messages_from_parent: {messages_from_parent}\n"
                f"pings_from_child: {pings_from_child}  ({parent_stats.num_received_by_topic[ping_from_child_topic]} - {start_pings_from_child})  on {ping_from_child_topic}\n"
                f"messages_from_child: {messages_from_child}\n"
                f"exp_pings_nominal: {exp_pings_nominal}\n"
                f"\n{h.summary_str()}\n"
            )
            assert (pings_from_child + pings_from_parent) >= exp_pings_nominal, err_str
            assert messages_from_child >= exp_pings_nominal, err_str
            assert messages_from_parent >= exp_pings_nominal, err_str

            # Test that ping not sent peridoically if messages are sent
            start_pings_from_parent = stats.num_received_by_topic[
                pings_from_parent_topic
            ]
            start_pings_from_child = parent_stats.num_received_by_topic[
                ping_from_child_topic
            ]
            start_messages_from_parent = stats.num_received
            start_messages_from_child = parent_stats.num_received
            reps = 50
            for _ in range(reps):
                parent.send_dbg(parent.downstream_client)
                await asyncio.sleep(0.01)
            pings_from_parent = (
                stats.num_received_by_topic[pings_from_parent_topic]
                - start_pings_from_parent
            )
            pings_from_child = (
                parent_stats.num_received_by_topic[ping_from_child_topic]
                - start_pings_from_child
            )
            messages_from_parent = stats.num_received - start_messages_from_parent
            messages_from_child = parent_stats.num_received - start_messages_from_child
            exp_pings_nominal = 2
            err_str = (
                f"\npings_from_parent: {pings_from_parent}  ({stats.num_received_by_topic[pings_from_parent_topic]} - {start_pings_from_parent})  on <{pings_from_parent_topic}>\n"
                f"messages_from_parent: {messages_from_parent}\n"
                f"pings_from_child: {pings_from_child}  ({parent_stats.num_received_by_topic[ping_from_child_topic]} - {start_pings_from_child})  on {ping_from_child_topic}\n"
                f"messages_from_child: {messages_from_child}\n"
                f"exp_pings_nominal: {exp_pings_nominal}\n"
                f"reps: {reps}, {reps * 0.5}\n"
                f"\n{h.summary_str()}\n"
            )
            assert pings_from_parent <= exp_pings_nominal, err_str
            assert pings_from_child <= exp_pings_nominal, err_str
            # Allow wide variance in number of messages exchanged - we are really testing pings, which
            # Should be should be close to 0 when a lot of messages are being exchanged.
            assert messages_from_parent >= reps * 0.5, err_str
            assert messages_from_child >= reps * 0.5, err_str

            parent.pause_acks()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                child.links.ack_manager.default_delay_seconds + 1,
                "ERROR waiting for for parent to be slow",
                err_str_f=h.summary_str,
            )
            parent.release_acks(clear=True)
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for parent to respond",
                err_str_f=h.summary_str,
            )
